"""
functions for loading and manipulating colors
"""
import json
import os


def colors_loading():
    """
    colors_loading allows Eurybia to load a json file which contains different
    palettes of colors that can be used in the plot
    Returns
    -------
    dict:
        contains all available pallets
    """
    current_path = os.path.dirname(os.path.abspath(__file__))
    jsonfile = os.path.join(current_path, "colors.json")
    with open(jsonfile) as openfile:
        colors_dic = json.load(openfile)
    return colors_dic


def select_palette(colors_dic, palette_name):
    """
    colors_loading allows Eurybia to load a json file which contains different
    palettes of colors that can be used in the plot
    Parameters
    ----------
    colors_dic : dict
        dictionnary with every palettes
    palette_name : String
        name of the palette
    Returns
    -------
    dict:
        contains colors of one palette
    """
    if palette_name not in colors_dic.keys():
        raise ValueError(f"Palette {palette_name} not found.")
    return colors_dic[palette_name]


def define_style(palette):
    """
    the define_style function is a function that uses a palette
    to define the different styles used in the different outputs
    of Eurybia
    Parameters
    ----------
    palette : dict
        contains colors of one palette
    Returns
    -------
    dict :
        contains different style elements
    """
    style_dict = dict()
    style_dict["dict_title"] = {
        "xanchor": "center",
        "yanchor": "middle",
        "x": 0.5,
        "y": 0.9,
        "font": {"size": 24, "family": "Arial", "color": palette["title_color"]},
    }

    featureimp_bar = palette["featureimp_bar"]
    style_dict["dict_featimp_colors"] = {
        "color": featureimp_bar,
        "line": {"color": palette["featureimp_bar"], "width": 0.5},
    }
    style_dict["featureimp_groups"] = list(palette["featureimp_groups"].values())

    style_dict["dict_xaxis_title"] = {"font": {"size": 16, "family": "Arial Black", "color": palette["axis_color"]}}
    style_dict["dict_yaxis_title"] = {"font": {"size": 16, "family": "Arial Black", "color": palette["axis_color"]}}
    style_dict["dict_xaxis"] = dict(
        linecolor="#BCCCDC",
        showspikes=True,  # Show spike line for X-axis
        # Format spike
        spikethickness=2,
        spikedash="dot",
        spikecolor="#999999",
        spikemode="across",
    )
    style_dict["dict_xaxis_continuous"] = {
        "font": {"size": 16, "family": "Arial Black", "color": palette["axis_color"]}
    }
    style_dict["dict_yaxis_continuous"] = {
        "font": {"size": 16, "family": "Arial Black", "color": palette["title_color"]},
        "text": "Density",
    }
    style_dict["dict_legend"] = {"title": ""}
    style_dict["template"] = "none"
    style_dict["height"] = 600
    style_dict["width"] = 900
    style_dict["univariate_cat_bar"] = palette["univariate_cat_bar"]
    style_dict["univariate_cont_bar"] = palette["univariate_cont_bar"]
    style_dict["datadrift_historical"] = palette["datadrift_historical"]
    style_dict["scatter_plot"] = palette["scatter_plot"]
    style_dict["scatter_line"] = palette["scatter_line"]
    style_dict["featimportance_colorscale"] = palette["featimportance_colorscale"]
    style_dict["contrib_colorscale"] = palette["contrib_colorscale"]
    style_dict["hovermode"] = "closest"
    return style_dict



"""
Report generation helper module.
"""
from base64 import b64encode
from datetime import datetime
from typing import Optional

import datapane as dp
import importlib_resources as ir
import pandas as pd
from jinja2 import Environment, FileSystemLoader
from shapash.explainer.smart_explainer import SmartExplainer

from eurybia import SmartDrift
from eurybia.report.project_report import DriftReport


def _load_custom_template(report: dp.Report) -> dp.Report:
    """
    This function feeds a customised html template to Datapane

    Parameters
    ----------
    report : datapane.Report
        Report object
    Returns
    ----------
    datapane.Report
    """
    report._local_writer.assets = ir.files("eurybia.report.assets")
    logo_img = (report._local_writer.assets / "logo_eurybia_dp.png").read_bytes()
    report._local_writer.logo = f"data:image/png;base64,{b64encode(logo_img).decode('ascii')}"
    template_loader = FileSystemLoader(report._local_writer.assets)
    template_env = Environment(loader=template_loader)
    template_env.globals["include_raw"] = dp.client.api.report.core.include_raw
    report._local_writer.template = template_env.get_template("report_template.html")
    return report


def _get_index(dr: DriftReport, project_info_file: str, config: Optional[dict]) -> dp.Page:
    """
    This function generates and returns a Datapane page containing the Eurybia report index

    Parameters
    ----------
    dr : DriftReport
        DriftReport object
    project_info_file : str
        Path to the file used to display some information about the project in the report.
    config : dict, optional
        Report configuration options.
    Returns
    ----------
    datapane.Page
    """

    eurybia_logo = """
    <html>
        <img src="https://eurybia.readthedocs.io/en/latest/_images/eurybia-fond-clair.png?raw=true" alt="drawing" width="150" height="150"/>
    </html>
    """

    # main block
    index_block = []

    # Title and logo
    index_block += [dp.Group(dp.HTML(eurybia_logo), dp.Text(f"# {dr.title_story}"), columns=2)]

    if config is not None and "title_description" in config.keys() and config["title_description"] != "":
        raw_title = config["title_description"]
        index_block += [dp.Text(f"## {raw_title}")]
    index_str = "## Eurybia Report contents  \n"

    # Tabs index
    if project_info_file is not None:
        index_str += "- Project information: report context and information  \n"
    index_str += "- Consistency Analysis: highlighting differences between the two datasets  \n"
    index_str += "- Data drift: In-depth data drift analysis \n"

    if dr.smartdrift.data_modeldrift is not None:
        index_str += "- Model drift: In-depth model drift analysis"

    index_block += [dp.Text(index_str)]

    # AUC
    auc_block = dr.smartdrift.plot.generate_indicator(
        fig_value=dr.smartdrift.auc, height=280, width=500, title="Datadrift classifier AUC"
    )

    # Jensen-Shannon
    if dr.smartdrift.deployed_model is not None:
        JS_block = dr.smartdrift.plot.generate_indicator(
            fig_value=dr.smartdrift.js_divergence,
            height=280,
            width=500,
            title="Jensen Shannon Datadrift",
            min_gauge=0,
            max_gauge=0.2,
        )
        index_block += [dp.Group(auc_block, JS_block, columns=3)]
    else:
        index_block += [dp.Group(auc_block, columns=2)]

    page_index = dp.Page(title="Index", blocks=index_block)
    return page_index


def _dict_to_text_blocks(text_dict, level=1):
    """
    This function recursively explores the dict and returns a Datapane Group containing other groups and text blocks fed with the dict
    Parameters
    ----------
    text_dict: dict
        This dict must contain string as keys, and dicts or strings as values
    level: int = 1
        Recursion level, starting at 1 to allow for direct string manipulation
    Returns
    ----------
    datapane.Group
        Group of blocks
    """
    blocks = []
    text = ""
    for k, v in text_dict.items():
        if isinstance(v, (str, int, float)) or v is None:
            if k.lower() == "date" and v.lower() == "auto":
                v = str(datetime.now())[:-7]
            text += f"**{k}** : {v}  \n"
        elif isinstance(v, dict):
            if text != "":
                blocks.append(dp.Text(text))
                text = ""
            blocks.append(
                dp.Group(dp.Text("#" * min(level, 6) + " " + str(k)), _dict_to_text_blocks(v, level + 1), columns=1)
            )
    if text != "":
        blocks.append(dp.Text(text))
    return dp.Group(blocks=blocks, columns=1)


def _get_project_info(dr: DriftReport) -> dp.Page:
    """
    This function generates and returns a Datapane page from a dict containing dicts and strings

    Parameters
    ----------
    dr : DriftReport
        DriftReport object

    Returns
    ----------
    datapane.Page
    """
    if dr.metadata is None:
        return None
    page_info = dp.Page(
        title="Project information",
        blocks=[_dict_to_text_blocks(dr.metadata)],
    )
    return page_info


def _get_consistency_analysis(dr: DriftReport) -> dp.Page:
    """
    This function generates and returns a Datapane page containing the Eurybia consistency analysis

    Parameters
    ----------
    dr : DriftReport
        DriftReport object

    Returns
    ----------
    datapane.Page
    """
    # Title
    blocks = [dp.Text("# Consistency Analysis")]

    # Manually ignored coluumns
    ignore_cols = pd.DataFrame({"ignore_cols": dr.smartdrift.ignore_cols}).rename(
        columns={"ignore_cols": "Ignored columns"}
    )
    blocks += [
        dp.Text("## Ignored columns in the report (manually excluded)"),
    ]
    if len(ignore_cols) > 0:
        blocks += [dp.Table(data=ignore_cols)]
    else:
        blocks += [dp.Text("- Ignored columns : None.")]

    # Column mismatches
    blocks += [
        dp.Text("## Consistency checks: column match between the 2 datasets."),
        dp.Text(
            """
            The columns identified in this section have been automatically removed from this analysis.
            Their presence would always be sufficient for the datadrift classifier to perfectly discriminate the two datasets (maximal data drift, AUC=1).
            """
        ),
    ]
    for k, v in dr.smartdrift.pb_cols.items():
        if len(v) > 0:
            blocks += [dp.Table(data=pd.DataFrame(v).transpose())]
        else:
            blocks += [dp.Text(f"- No {k.lower()} have been detected.")]

    blocks += [
        dp.Text("### Unique values identified:"),
        dp.Text(
            """
            This section displays categorical features in which unique values differ.
            This analysis has been performed on unstratified samples of the baseline and current datasets.
            Missing or added unique values can be caused by this sampling.
            Columns identified in this section have been kept for the analysis.
            """
        ),
    ]
    if len(dr.smartdrift.err_mods) > 0:
        blocks += [
            dp.Table(
                data=pd.DataFrame(dr.smartdrift.err_mods)
                .rename(columns={"err_mods": "Modalities present in one dataset and absent in the other :"})
                .transpose()
            )
        ]
    else:
        blocks += [dp.Text("- No modalities have been detected as present in one dataset and absent in the other.")]

    page_consistency = dp.Page(title="Consistency Analysis", blocks=blocks)
    return page_consistency


def _get_datadrift(dr: DriftReport) -> dp.Page:
    """
    This function generates and returns a Datapane page containing the Eurybia data drift analysis

    Parameters
    ----------
    dr : DriftReport
        DriftReport object

    Returns
    ----------
    datapane.Page
    """
    # Loop for save in list plots of display analysis
    plot_dataset_analysis = []
    table_dataset_analysis = []
    fig_list, labels, table_list = dr.display_dataset_analysis(global_analysis=False)["univariate"]
    for i in range(len(labels)):
        plot_dataset_analysis.append(dp.Plot(fig_list[i], label=labels[i]))
        table_dataset_analysis.append(dp.Table(table_list[i], label=labels[i]))

    # Loop for save in list plots of display analysis
    plot_datadrift_contribution = []
    fig_list, labels = dr.display_model_contribution()
    for i in range(len(labels)):
        plot_datadrift_contribution.append(dp.Plot(fig_list[i], label=labels[i]))
    blocks = [
        dp.Text("# Data drift"),
        dp.Text(
            """The data drift detection methodology is based on the ability of a model classifier to identify whether
        a sample belongs to one or another dataset.
        For this purpose a target (0) is assigned to the baseline dataset and a second target (1) to the current dataset.
        A classification model (catboost) is trained to predict this target.
        As such, the data drift classifier performance is directly related to the difference between two datasets.
        A marked difference will lead to an easy classification (final AUC close to 1).
        Oppositely, highly similars datasets will lead to poor data drift classifier performance (final AUC close to 0.5)."""
        ),
        dp.Text("## Detecting data drift"),
        dp.Text("### Datadrift classifier model perfomances"),
        dp.Text(
            """The closer your AUC is from 0.5 the less your data drifted.
             The closer your AUC is from 1 the more your data drifted"""
        ),
        dp.Plot(
            dr.smartdrift.plot.generate_indicator(
                fig_value=dr.smartdrift.auc, height=300, width=500, title="Datadrift classifier AUC"
            )
        ),
        dp.Text("## Importance of features in data drift"),
        dp.Text("### Global feature importance plot"),
        dp.Text(
            """Bar chart representing the feature importance of each feature for the datadrift classifier.
             This parameter is a direct measure of the importance of a feature to perform the classification."""
        ),
        dp.Plot(dr.explainer.plot.features_importance()),
    ]
    if dr.smartdrift.deployed_model is not None:
        blocks += [
            dp.Text("### Feature importance overview"),
            dp.Text(
                """Scatter plot depicting, for each feature, the feature importance of the deployed model as a function of the datadrift classifier
                 feature importance. This graph thus highlight the real importance of a data drift for the deployed model classification.
                Interpretation based on graphical feature location:
            - Top left : Feature highly important for the deployed model and with low data drift
            - Bottom left : Feature with moderated importance for the deployed model and with low data drift
            - Bottom right : Feature with moderated importance for the deployed model but with high data drift.
            This feature might require your attention.
            - Top right : Feature highly important for the deployed model and high drift. This feature requires your attention.
            """
            ),
            dp.Plot(dr.smartdrift.plot.scatter_feature_importance()),
        ]
    blocks += [
        dp.Text("## Dataset analysis"),
        dp.Text(
            """This section provides numerical and graphical analysis of the 2 datasets distributions,
            making easier the study of the most important variable for drift detection."""
        ),
        dp.Text("### Global analysis"),
        dp.Table(dr._display_dataset_analysis_global()),
        dp.Text("### Univariate analysis"),
        dp.Text(
            """Bar chart showing the unique values distribution of a feature.
        Using the drop-down menu, it is possible to select the feature of interest.
        Features are sorted according to their respective importance in the datadrift classifier.
        For categorical features, the possible values are sorted by descending difference between the two datasets."""
        ),
        dp.Select(blocks=plot_dataset_analysis),
        dp.Select(blocks=table_dataset_analysis),
    ]
    if dr.smartdrift.deployed_model is not None:
        blocks += [
            dp.Text("### Distribution of predicted values"),
            dp.Text(
                "Histogram density showing the distributions of the production model outputs on both baseline and current datasets."
            ),
            dp.Plot(
                dr.smartdrift.plot.generate_fig_univariate(df_all=dr.smartdrift.df_predict, col="Score", hue="dataset")
            ),
            dp.Text(
                """Jensen Shannon Divergence (JSD). The JSD measures the effect of a data drift on the deployed model performance.
            A value close to 0 indicates similar data distributions, while a value close to 1 tend to indicate distinct data distributions
            with a negative effect on the deployed model performance."""
            ),
            dr.smartdrift.plot.generate_indicator(
                fig_value=dr.smartdrift.js_divergence,
                height=280,
                width=500,
                title="Jensen Shannon Datadrift",
                min_gauge=0,
                max_gauge=0.2,
            ),
        ]
    blocks += [
        dp.Text("## Feature contribution on data drift's detection"),
        dp.Text(
            """This graph represents the contribution of a variable to the data drift detection.
        This representation constitutes a support to understand the drift when the analysis of the dataset is unclear.
        In the drop-down menu, features are sorted by importance in the data drift detection."""
        ),
        dp.Select(blocks=plot_datadrift_contribution),
    ]
    if dr.smartdrift.historical_auc is not None:
        blocks += [
            dp.Text("## Historical Data drift"),
            dp.Text(
                "Line chart showing the metrics evolution of the datadrift classifier over the given period of time."
            ),
            dp.Plot(dr.smartdrift.plot.generate_historical_datadrift_metric()),
        ]
    page_datadrift = dp.Page(title="Data drift", blocks=blocks)
    return page_datadrift


def _get_modeldrift(dr: DriftReport) -> dp.Page:
    """
    This function generates and returns a Datapane page containing the Eurybia model drift analysis

    Parameters
    ----------
    dr : DriftReport
        DriftReport object

    Returns
    ----------
    datapane.Page
    """
    # Loop for save in list plots of display model drift
    if dr.smartdrift.data_modeldrift is not None:
        plot_modeldrift = []
        fig_list, labels = dr.display_data_modeldrift()
        if labels == []:
            plot_modeldrift = dp.Plot(fig_list[0])
            modeldrift_plot = plot_modeldrift
        else:
            for i in range(len(labels)):
                plot_modeldrift.append(dp.Plot(fig_list[i], label=labels[i]))
            modeldrift_plot = dp.Select(blocks=plot_modeldrift, label="reference_columns")
    else:
        modeldrift_plot = dp.Text("## Smartdrift.data_modeldrift is None")
    blocks = [
        dp.Text("# Model drift"),
        dp.Text(
            """This section provides support to monitor the production model's performance over time.
    This requires the performance history as input."""
        ),
        dp.Text("## Performance evolution of the deployed model"),
        dp.Text("Line chart of deployed model performances as a function of time"),
        modeldrift_plot,
    ]
    page_modeldrift = dp.Page(title="Model drift", blocks=blocks)
    return page_modeldrift


def execute_report(
    smartdrift: SmartDrift,
    explainer: SmartExplainer,
    project_info_file: str,
    output_file: str,
    config: Optional[dict] = None,
):
    """
    Creates the report

    Parameters
    ----------
    smartdrift : eurybia.core.smartdrift.SmartDrift object
        Compiled SmartDrift class
    explainer : shapash.explainer.smart_explainer.SmartExplainer object
        Compiled shapash explainer.
    project_info_file : str
        Path to the file used to display some information about the project in the report.
    config : dict, optional
        Report configuration options.
    output_file : str
            Path to the HTML file to write
    """

    if config is None:
        config = {}

    dr = DriftReport(
        smartdrift=smartdrift,
        explainer=explainer,  # rename to match kwarg
        project_info_file=project_info_file,
        config=config,
    )

    pages = []
    pages.append(_get_index(dr, project_info_file, config))
    if project_info_file is not None:
        pages.append(_get_project_info(dr))
    pages.append(_get_consistency_analysis(dr))
    pages.append(_get_datadrift(dr))
    if dr.smartdrift.data_modeldrift is not None:
        pages.append(_get_modeldrift(dr))

    report = dp.Report(blocks=pages)
    report = _load_custom_template(report)
    report._save(
        path=output_file, open=False, formatting=dp.ReportFormatting(light_prose=False, width=dp.ReportWidth.MEDIUM)
    )



"""
Unit test utils.py
"""
import os
import unittest
from pathlib import Path

import pandas as pd

from eurybia.utils.utils import base_100, convert_string_to_int_keys, get_project_root, round_to_k, truncate_str


class Testutils(unittest.TestCase):
    """
    Unit test utils.py
    """

    def test_base_100(self):
        """
        Unit test base_100 function
        """
        seriestest = pd.Series([2, 8])
        resulttest = base_100(seriestest)

        assert isinstance(resulttest, pd.Series)
        pd.testing.assert_series_equal(resulttest, pd.Series([0.2, 0.8]))

    def test_truncate_str_1(self):
        """
        unit test truncate_str()
        """
        trunc_res = truncate_str(12)
        assert trunc_res == 12

    def test_truncate_str_2(self):
        """
        unit test truncate_str()
        """
        trunc_res = truncate_str("this is a test", 50)
        assert trunc_res == "this is a test"

    def test_truncate_str_3(self):
        """
        unit test truncate_str()
        """
        trunc_res = truncate_str("this is a test", 10)
        assert trunc_res == "this is a..."

    def test_convert_string_to_int_keys(self):
        """
        unit test convert_string_to_int_keys()
        """
        res = convert_string_to_int_keys({"0": 0, "1": 1})
        assert all(isinstance(k, int) for k in res)

    def test_convert_string_to_int_keys_invalid(self):
        """
        unit test convert_string_to_int_keys()
        """
        self.assertRaises(ValueError, convert_string_to_int_keys, {"foo": 0})

    def test_get_project_root_samedir(self):
        """
        unit test get_project_root()
        tests that get_project_root returns the same directory whether it is called from here or one directory up
        """
        project_root1 = get_project_root()
        test_dir = Path(__file__).parent.resolve()
        os.chdir("../")
        assert get_project_root() == project_root1
        os.chdir(test_dir)

    def test_round_to_k(self):
        """
        unit test round_to_k()
        """
        res1 = round_to_k(2.22, 2)
        res2 = round_to_k(2.22222222, 2)
        res3 = round_to_k(2.22222222, 5)
        res4 = round_to_k(2.2, 1)
        res5 = round_to_k(3, 1)
        res6 = round_to_k(3.0, 1)
        res7 = round_to_k(3.0000, 5)
        should_be_floats = [res1, res2, res3]
        should_be_ints = [res4, res5, res6, res7]

        assert len(str(res1)) == 3  # the dot counts, hence 2 figures + 1 dot = 3 chars
        assert len(str(res2)) == 3
        assert len(str(res3)) == 6  # the dot counts, hence 5 figures + 1 dot = 6 chars
        assert len(str(res4)) == 1
        assert len(str(res5)) == 1
        assert len(str(res6)) == 1
        assert len(str(res7)) == 1

        assert all(isinstance(r, float) for r in should_be_floats)
        assert all(isinstance(r, int) for r in should_be_ints)



"""
Smart plotter module
"""

import copy

# ----- Eurybia packages
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.figure_factory as ff
import seaborn as sns
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from eurybia.report.common import VarType, compute_col_types
from eurybia.style.style_utils import colors_loading, define_style, select_palette


class SmartPlotter:
    """
    The smartplotter class includes all the methods used to display graphics

    Each SmartPlotter method is easy to use from a Smart Drift object,
    just use the following syntax

    Attributes
    ----------
    smartdrift: object
        SmartDrift object
    _palette_name : str (default: 'eurybia')
        Name of the palette used for the colors of the report (refer to style folder).
    _style_dict: dict
            Dict contains dicts of the colors used in the different plots
    Example
    --------
    >>> SD = Smartdrift()
    >>> SD.compile()
    >>> SD.plot.my_plot_method(param=value)

    """

    def __init__(self, smartdrift):
        self._palette_name = list(colors_loading().keys())[0]
        self._style_dict = define_style(select_palette(colors_loading(), self._palette_name))
        self.smartdrift = smartdrift

    def generate_fig_univariate(
        self,
        col: str,
        hue: Optional[str] = None,
        df_all: Optional[pd.DataFrame] = None,
        dict_color_palette: Optional[dict] = None,
    ) -> plt.Figure:
        """
        Returns a plotly figure containing the distribution of any kind of feature
        (continuous, categorical).

        If the feature is categorical and contains too many categories, the smallest
        categories are grouped into a new 'Other' category so that the graph remains
        readable.

        The input dataframe should contain the column of interest and a column that is used
        to distinguish two types of values (ex. 'train' and 'test')

        Parameters
        ----------
        df_all : pd.DataFrame
            The input dataframe that contains the column of interest
        col : str
            The column of interest
        hue : str
            The column used to distinguish the values (ex. 'train' and 'test')
        type: str
            The type of the series ('continous' or 'categorical')

        Returns
        -------
        plotly.graph_objs._figure.Figure
        """
        if hue is None:
            hue = self.smartdrift._datadrift_target
        if df_all is None:
            df_all = self.smartdrift._df_concat
            df_all.loc[df_all[hue] == 0, hue] = list(self.smartdrift.dataset_names.keys())[1]
            df_all.loc[df_all[hue] == 1, hue] = list(self.smartdrift.dataset_names.keys())[0]
        if dict_color_palette is None:
            dict_color_palette = self._style_dict
        col_types = compute_col_types(df_all=df_all)
        if col_types[col] == VarType.TYPE_NUM:
            fig = self.generate_fig_univariate_continuous(df_all, col, hue=hue, dict_color_palette=dict_color_palette)
        elif col_types[col] == VarType.TYPE_CAT:
            fig = self.generate_fig_univariate_categorical(df_all, col, hue=hue, dict_color_palette=dict_color_palette)
        else:
            raise NotImplementedError("Series dtype not supported")
        return fig

    def generate_fig_univariate_continuous(
        self,
        df_all: pd.DataFrame,
        col: str,
        hue: str,
        dict_color_palette: dict,
        template: Optional[str] = None,
        title: Optional[str] = None,
        xaxis_title: Optional[dict] = None,
        yaxis_title: Optional[dict] = None,
        xaxis: Optional[str] = None,
        height: Optional[str] = None,
        width: Optional[str] = None,
        hovermode: Optional[str] = None,
    ) -> plotly.graph_objs._figure.Figure:

        """
        Returns a plotly figure containing the distribution of a continuous feature.

        Parameters
        ----------
        df_all : pd.DataFrame
            The input dataframe that contains the column of interest
        col : str
            The column of interest
        hue : str
            The column used to distinguish the values (ex. 'train' and 'test')
        template: str, , optional
            Template (background style) for the plot
        title: str, optional
            Plot title
        xaxis_title: str, , optional
            X axis title
        yaxis_title: str, , optional
            y axis title
        xaxis: str, , optional
            X axis options (spike line, margin, range ...)
        height: str, , optional
            Height of the plot
        width: str, , optional
            Width of the plot
        hovermode: str,n , optional
            Type of labels displaying on mouse hovering
        Returns
        -------
        plotly.graph_objs._figure.Figure
        """
        df_all.loc[:, col].fillna(0, inplace=True)
        datasets = [df_all[df_all[hue] == val][col].values.tolist() for val in df_all[hue].unique()]

        fig = ff.create_distplot(
            datasets,
            group_labels=[str(val) for val in df_all[hue].unique()],
            colors=list(self._style_dict["univariate_cont_bar"].values()),
            show_hist=False,
            show_curve=True,
            show_rug=False,
        )
        if template is None:
            template = self._style_dict["template"]
        if title is None:
            title = self._style_dict["dict_title"]
        if xaxis_title is None:
            xaxis_title = self._style_dict["dict_xaxis_continuous"]
            xaxis_title["text"] = col
        if yaxis_title is None:
            yaxis_title = self._style_dict["dict_yaxis_continuous"]
        if xaxis is None:
            xaxis = self._style_dict["dict_xaxis"]
        if height is None:
            height = self._style_dict["height"]
        if width is None:
            width = self._style_dict["width"]
        if hovermode is None:
            hovermode = self._style_dict["hovermode"]

        fig.update_layout(
            template=template,
            title=title,
            xaxis_title=xaxis_title,
            yaxis_title=yaxis_title,
            xaxis=xaxis,
            height=height,
            width=width,
            hovermode=hovermode,
        )
        fig.update_traces(hovertemplate="%{y:.2f}", showlegend=True)
        fig.update_xaxes(showgrid=False)
        fig.update_yaxes(showgrid=False)

        return fig

    def generate_fig_univariate_categorical(
        self,
        df_all: pd.DataFrame,
        col: str,
        hue: str,
        dict_color_palette: dict,
        nb_cat_max: int = 15,
        template: Optional[str] = None,
        title: Optional[str] = None,
        xaxis_title: Optional[dict] = None,
        yaxis_title: Optional[dict] = None,
        xaxis: Optional[str] = None,
        height: Optional[str] = None,
        width: Optional[str] = None,
        hovermode: Optional[str] = None,
        legend: Optional[str] = None,
    ) -> plotly.graph_objs._figure.Figure:
        """
        Returns a plotly figure containing the distribution of a categorical feature.

        If the feature is categorical and contains too many categories, the smallest
        categories are grouped into a new 'Other' category so that the graph remains
        readable.

        Parameters
        ----------
        df_all : pd.DataFrame
            The input dataframe that contains the column of interest
        col : str
            The column of interest
        hue : str
            The column used to distinguish the values (ex. 'train' and 'test')
        nb_cat_max : int
            The number max of categories to be displayed. If the number of categories
            is greater than nb_cat_max then groups smallest categories into a new
            'Other' category
        template: str, optional
            Template (background style) for the plot
        title: str, optional
            Plot title
        xaxis_title: str, optional
            X axis title
        yaxis_title: str, optional
            y axis title
        xaxis: str, optional
            X axis options (spike line, margin, range ...)
        height: str, optional
            Height of the plot
        width: str, optional
            Width of the plot
        hovermode: str, optional
            Type of labels displaying on mouse hovering
        legend: str, optional
            Axis legends
        Returns
        -------
        plotly.graph_objs._figure.Figure
        """
        df_cat = df_all.groupby([col, hue]).agg({col: "count"}).rename(columns={col: "count"}).reset_index()
        df_cat["Percent"] = df_cat["count"] * 100 / df_cat.groupby(hue)["count"].transform("sum")

        if pd.api.types.is_numeric_dtype(df_cat[col].dtype):
            df_cat = df_cat.sort_values(col, ascending=True)
            df_cat[col] = df_cat[col].astype(str)

        nb_cat = df_cat.groupby([col]).agg({"count": "sum"}).reset_index()[col].nunique()

        if nb_cat > nb_cat_max:
            df_cat = self._merge_small_categories(df_cat=df_cat, col=col, hue=hue, nb_cat_max=nb_cat_max)

        df_to_sort = df_cat.copy().reset_index(drop=True)
        df_to_sort["Sorted_indicator"] = df_to_sort.sort_values([col]).groupby([col])["Percent"].diff()
        df_to_sort["Sorted_indicator"] = np.abs(df_to_sort["Sorted_indicator"])
        df_sorted = df_to_sort.dropna()[[col, "Sorted_indicator"]]

        df_cat = (
            df_cat.merge(df_sorted, how="left", on=col)
            .sort_values("Sorted_indicator", ascending=True)
            .drop("Sorted_indicator", axis=1)
        )

        df_cat["Percent_displayed"] = df_cat["Percent"].apply(lambda row: str(round(row, 2)) + " %")

        modalities = df_cat[hue].unique().tolist()

        fig1 = px.bar(
            df_cat[df_cat[hue] == modalities[0]],
            x="Percent",
            y=col,
            orientation="h",
            barmode="group",
            color=hue,
            text="Percent_displayed",
        )
        fig1.update_traces(marker_color=list(self._style_dict["univariate_cat_bar"].values())[0], showlegend=True)

        fig2 = px.bar(
            df_cat[df_cat[hue] == modalities[1]],
            x="Percent",
            y=col,
            orientation="h",
            barmode="group",
            color=hue,
            text="Percent_displayed",
        )
        fig2.update_traces(marker_color=list(self._style_dict["univariate_cat_bar"].values())[1], showlegend=True)

        fig = fig1.add_trace(fig2.data[0])

        fig.update_xaxes(showgrid=False, showticklabels=True)
        fig.update_yaxes(showgrid=False, showticklabels=True, automargin=True)
        fig.update_traces(showlegend=True, textposition="outside", cliponaxis=False)

        if template is None:
            template = self._style_dict["template"]
        if title is None:
            title = self._style_dict["dict_title"]
        if xaxis_title is None:
            xaxis_title = self._style_dict["dict_xaxis_title"]
        if yaxis_title is None:
            yaxis_title = self._style_dict["dict_yaxis_title"]
            yaxis_title["text"] = col
        if height is None:
            height = self._style_dict["height"]
        if width is None:
            width = self._style_dict["width"]
        if hovermode is None:
            hovermode = self._style_dict["hovermode"]
        if legend is None:
            legend = self._style_dict["dict_legend"]

        fig.update_layout(
            template=template,
            title=title,
            xaxis_title=xaxis_title,
            height=height,
            width=width,
            yaxis_title=yaxis_title,
            hovermode=hovermode,
            legend=legend,
            xaxis_range=[0, max(df_cat["Percent"]) + 10],
        )

        return fig

    def _merge_small_categories(self, df_cat: pd.DataFrame, col: str, hue: str, nb_cat_max: int) -> pd.DataFrame:
        """
        Merges categories of column 'col' of df_cat into 'Other' category so that
        the number of categories is less than nb_cat_max.
        """
        df_cat_sum_hue = df_cat.groupby([col]).agg({"count": "sum"}).reset_index()
        list_cat_to_merge = df_cat_sum_hue.sort_values("count", ascending=False)[col].to_list()[nb_cat_max - 1 :]
        df_cat_other = (
            df_cat.loc[df_cat[col].isin(list_cat_to_merge)].groupby(hue, as_index=False)[["count", "Percent"]].sum()
        )
        df_cat_other[col] = "Other"
        return pd.concat([df_cat.loc[~df_cat[col].isin(list_cat_to_merge)], df_cat_other])

    def scatter_feature_importance(
        self, feature_importance: pd.DataFrame = None, datadrift_stat_test: pd.DataFrame = None
    ) -> plotly.graph_objs._figure.Figure:
        """
        Displays scatter of feature importance between drift
        model and production one extracted from a datasets created
        during the compile step.

        Parameters
        ----------
        feature_importance : pd.DataFrame, optional
            DataFrame containing feature importance for each features from production and drift model.
        datadrift_stat_test: pd.DataFrame, optional
            DataFrame containing the result of datadrift univariate tests
        Returns
        -------
        plotly.express.scatter
        """
        dict_t = copy.deepcopy(self._style_dict["dict_title"])
        dict_xaxis = copy.deepcopy(self._style_dict["dict_xaxis_title"])
        dict_yaxis = copy.deepcopy(self._style_dict["dict_yaxis_title"])
        title = "<b>Datadrift Vs Feature Importance</b>"
        dict_t["text"] = title
        dict_xaxis["text"] = "Datadrift Importance"
        dict_yaxis["text"] = "Feature Importance - Deployed Model"

        if feature_importance is None:
            feature_importance = self.smartdrift.feature_importance.set_index("feature")
        if datadrift_stat_test is None:
            datadrift_stat_test = self.smartdrift.datadrift_stat_test

        data = datadrift_stat_test.join(feature_importance)
        data["features"] = data.index
        # symbols
        stat_test_list = list(data["testname"].unique())
        symbol_list = [0, 13]
        symbol_dict = dict(zip(stat_test_list, symbol_list))

        hv_text = [
            f"<b>Feature: {feat}</b><br />Deployed Model Importance: {depimp*100:.1f}%<br />"
            + f"Datadrift test: {t} - pvalue: {pv:.5f}<br />"
            + f"Datadrift model Importance: {ddrimp*100:.1f}"
            for feat, depimp, t, pv, ddrimp in zip(
                *map(data.get, ["features", "deployed_model", "testname", "pvalue", "datadrift_classifier"])
            )
        ]

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=data["datadrift_classifier"],
                y=data["deployed_model"],
                marker_symbol=datadrift_stat_test["testname"].apply(lambda x: symbol_dict[x]),
                mode="markers",
                showlegend=False,
                hovertext=hv_text,
                hovertemplate="%{hovertext}<extra></extra>",
            )
        )

        fig.update_traces(marker={"size": 15, "opacity": 0.8, "line": {"width": 0.8, "color": "white"}})

        fig.data[0].marker.color = data["pvalue"]
        fig.data[0].marker.coloraxis = "coloraxis"
        fig.layout.coloraxis.colorscale = self._style_dict["featimportance_colorscale"]
        fig.layout.coloraxis.colorbar = {"title": {"text": "Univariate<br />DataDrift Test<br />Pvalue"}}

        height = self._style_dict["height"]
        width = self._style_dict["width"]
        hovermode = self._style_dict["hovermode"]
        template = self._style_dict["template"]

        fig.update_layout(
            template=template,
            title=dict_t,
            xaxis_title=dict_xaxis,
            yaxis_title=dict_yaxis,
            height=height,
            width=width,
            hovermode=hovermode,
        )

        return fig

    def generate_historical_datadrift_metric(
        self,
        datadrift_historical: pd.DataFrame = None,
        template: Optional[str] = None,
        title: Optional[str] = None,
        xaxis_title: Optional[str] = None,
        yaxis_title: Optional[str] = None,
        xaxis: Optional[str] = None,
        height: Optional[str] = None,
        width: Optional[str] = None,
        hovermode: Optional[str] = None,
    ) -> plotly.graph_objs._figure.Figure:
        """
        Displays line plot of the evolution of the datadrift metrics :
        AUC of Datadrift classifier and if deployed_model fill, Jensen Shannon divergence of distribution of prediction
        Parameters
        ----------
        datadrift_historical : pd.DataFrame
            DataFrame with date, datadrif classifer auc and jensen shannon prediction divergence if deployed_model fill
        template: str, optional
            Template (background style) for the plot
        title: str, optional
            Plot title
        xaxis_title: str, optional
            X axis title
        yaxis_title: str, optional
            y axis title
        xaxis: str, optional
            X axis options (spike line, margin, range ...)
        height: str, optional
            Height of the plot
        width: str, optional
            Width of the plot
        hovermode: str, optional
            Type of labels displaying on mouse hovering
        Returns
        -------
        plotly.express.line
        """
        if datadrift_historical is None:
            datadrift_historical = self.smartdrift.historical_auc
        if datadrift_historical is not None:
            if self.smartdrift.deployed_model is not None:
                datadrift_historical = datadrift_historical[["date", "auc", "JS_predict"]]
                datadrift_historical = (
                    datadrift_historical.groupby(["date"])[["auc", "JS_predict"]].mean().reset_index()
                )
                datadrift_historical.sort_values(by="date", inplace=True)
            else:
                datadrift_historical = datadrift_historical[["date", "auc"]]
                datadrift_historical = datadrift_historical.groupby("date")["auc"].mean().reset_index()
                datadrift_historical.sort_values(by="date", inplace=True)

            datadrift_historical["auc_displayed"] = datadrift_historical["auc"].round(2)

            if self.smartdrift.deployed_model is not None:

                fig = make_subplots(specs=[[{"secondary_y": True}]])
                fig.add_trace(
                    go.Scatter(
                        x=datadrift_historical["date"], y=datadrift_historical["auc"], name="Datadrift classifier AUC"
                    ),
                    secondary_y=False,
                )

                fig.add_trace(
                    go.Scatter(
                        x=datadrift_historical["date"],
                        y=datadrift_historical["JS_predict"],
                        name="Jensen_Shannon Prediction Divergence",
                    ),
                    secondary_y=True,
                )

                fig.update_layout(title_text="Evolution of data drift")
                fig.update_yaxes(title_text="<b>Datadrift classifier AUC</b>  ", secondary_y=False)
                fig.update_yaxes(title_text="<b>Jensen_Shannon Prediction Divergence</b> ", secondary_y=True)
                fig.update_yaxes(range=[0.5, 1], secondary_y=False)
                fig.update_yaxes(range=[0, 0.3], secondary_y=True)
            else:
                fig = px.line(
                    datadrift_historical,
                    x="date",
                    y="auc",
                    title="AUC's Evolution of Datadrift classifier",
                    text="auc_displayed",
                )
                fig.update_yaxes(title_text="<b>Datadrift classifier AUC</b>")
                fig.update_yaxes(range=[0.5, 1])

            fig.update_traces(textposition="bottom right")

            if template is None:
                template = self._style_dict["template"]
            if title is None:
                title = self._style_dict["dict_title"]
            if xaxis_title is None:
                xaxis_title = self._style_dict["dict_xaxis_title"]
            if height is None:
                height = self._style_dict["height"]
            if width is None:
                width = self._style_dict["width"]
            if hovermode is None:
                hovermode = self._style_dict["hovermode"]

            fig.update_xaxes(showgrid=False)
            fig.update_layout(
                template=template,
                title=title,
                xaxis_title=xaxis_title,
                height=height,
                width=width,
                hovermode=hovermode,
            )
            return fig

    def generate_modeldrift_data(
        self,
        data_modeldrift: pd.DataFrame = None,
        metric: str = "performance",
        reference_columns: list = list(),
        template: Optional[str] = None,
        title: Optional[str] = None,
        xaxis_title: Optional[str] = None,
        yaxis_title: Optional[str] = None,
        xaxis: Optional[str] = None,
        height: Optional[str] = None,
        width: Optional[str] = None,
        hovermode: Optional[str] = None,
    ) -> plotly.graph_objs._figure.Figure:
        """
        Displays line plot of the evolution of the Lift computed for deployed model with several criterias.

        Parameters
        ----------
        data_modeldrift : pd.DataFrame
            DataFrame containing the aggregated informations to display modeldrift.
        metric : str
            Column name of the metric computed
        reference_columns : list
            list of reference columns used to display the metric according to different criteria
        title: str, optional
            Plot title
        xaxis_title: str, optional
            X axis title
        yaxis_title: str, optional
            y axis title
        xaxis: str, optional
            X axis options (spike line, margin, range ...)
        height: str, optional
            Height of the plot
        width: str, optional
            Width of the plot
        hovermode: str, optional
            Type of labels displaying on mouse hovering
        Returns
        -------
        plotly.express.line
        """
        if data_modeldrift is None:
            data_modeldrift = self.smartdrift.data_modeldrift
            if data_modeldrift is None:
                raise Exception(
                    """You should run the add_data_modeldrift method before displaying model drift performances.
                For more information see the documentation"""
                )
        data_modeldrift[metric] = data_modeldrift[metric].apply(
            lambda row: round(row, len([char for char in str(row).split(".")[1] if char == "0"]) + 3)
        )

        fig = px.line(
            data_modeldrift,
            x="Date",
            y=metric,
            hover_data=reference_columns,
            title="Performance's Evolution on deployed model",
            text=metric,
        )

        fig.update_traces(textposition="top right")

        if template is None:
            template = self._style_dict["template"]
        if title is None:
            title = self._style_dict["dict_title"]
        if xaxis_title is None:
            xaxis_title = self._style_dict["dict_xaxis_title"]
        if yaxis_title is None:
            yaxis_title = self._style_dict["dict_yaxis_continuous"]
        if height is None:
            height = self._style_dict["height"]
        if width is None:
            width = self._style_dict["width"]
        if hovermode is None:
            hovermode = self._style_dict["hovermode"]

        fig.update_xaxes(showgrid=False)
        fig.update_layout(
            template=template,
            title=title,
            xaxis_title=xaxis_title,
            height=height,
            width=width,
            yaxis_title=yaxis_title,
            hovermode=hovermode,
        )

        fig.data[0].line.color = self._style_dict["datadrift_historical"]
        fig.data[-1].marker.color = self._style_dict["datadrift_historical"]

        return fig

    def define_style_attributes(self, colors_dict):
        """
        define_style_attributes allows Eurybia user to change the color of plot
        Parameters
        ----------
        colors_dict: dict
            Dict of the colors used in the different plots
        """
        self._style_dict = define_style(colors_dict)

        if hasattr(self, "pred_colorscale"):
            delattr(self, "pred_colorscale")

    def generate_indicator(
        self,
        fig_value: float,
        min_gauge: float = 0.5,
        max_gauge: float = 1,
        height: Optional[float] = 300,
        width: Optional[float] = 500,
        title: Optional[str] = "Metric",
    ) -> plotly.graph_objs._figure.Figure:
        """
        Displays an indicator in a colorbar
        Parameters
        ----------
        fig_value: float
            Value to display on figure
        min_gauge: float, (default: 0.5)
            range min in gauge display
        max_gauge: float, (default: 1)
            range max in gauge display
        height: str, optional
            Height of the plot
        width: str, optional
            Width of the plot
        title: str, optional
            Plot title
        """
        color = sns.blend_palette(["green", "yellow", "orange", "red"], 100)
        color = color.as_hex()
        list_color_glob = list()
        threshold = [i for i in np.arange(min_gauge, max_gauge, (max_gauge - min_gauge) / len(color))]
        for i in range(1, len(threshold) + 1):
            dict_color = dict()
            if i == len(threshold):
                rang = [threshold[i - 1], 1]
            else:
                rang = [threshold[i - 1], threshold[i]]
            dict_color["range"] = rang
            dict_color["color"] = color[i - 1]
            list_color_glob.append(dict_color)
        fig = go.Figure(
            go.Indicator(
                mode="gauge+number",
                value=round(fig_value, 2),
                domain={"x": [0, 1], "y": [0, 1]},
                title={"text": title, "align": "center", "font": {"size": 20}},
                gauge={
                    "axis": {"range": [min_gauge, max_gauge], "ticktext": ["No Drift", "High Drift"], "tickwidth": 1},
                    "bar": {"color": "black"},
                    "borderwidth": 0,
                    "steps": list_color_glob,
                },
            )
        )
        fig.update_layout(
            height=height,
            width=width,
        )

        return fig



"""
Module used in the base_report notebook to generate report
"""
import copy
import logging
import os
from typing import Dict, Optional, Union

import jinja2
import pandas as pd
from shapash.explainer.smart_explainer import SmartExplainer

from eurybia.core.smartdrift import SmartDrift
from eurybia.report.common import compute_col_types
from eurybia.report.data_analysis import perform_global_dataframe_analysis, perform_univariate_dataframe_analysis
from eurybia.utils.io import load_yml
from eurybia.utils.utils import get_project_root

logging.basicConfig(level=logging.INFO)

template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(get_project_root(), "report", "html"))
template_env = jinja2.Environment(loader=template_loader)


dict_font = dict(family="Arial Black", size=18)


class DriftReport:
    """
    The DriftReport class allows to generate compare two datasets.
    One used to train the model the other build for production purposes.
    It analyzes the data and the model used in order to provide interesting
    insights that can be shared with non technical person.

    Attributes
    ----------
    smartdrift: object
            SmartDrift object
    explainer : shapash.explainer.smart_explainer.SmartExplainer
         A shapash SmartExplainer object that has already be compiled
    title_story : str
            Report title
    metadata : dict
        Information about the project (author, description, ...)
    df_predict : pd.DataFrame
        Dataframe of predicted values computed on both df_baseline and df_current
    feature_importance : pd.DataFrame, optional (default: None)
        Dataframe of feature importance from production model and drift model
    config : dict, optional
        Configuration options for the report
    """

    def __init__(
        self,
        smartdrift: SmartDrift,
        explainer: SmartExplainer,
        project_info_file: Optional[str] = None,
        config: Optional[Dict] = None,
    ):
        """
        Parameters
        ----------
        smartdrift: object
            SmartDrift object
        explainer : shapash.explainer.smart_explainer.SmartExplainer
            A shapash SmartExplainer object that has already be compiled
        project_info_file : str
            Path to the yml file containing information about the project (author, description, ...)
        config : dict, optional
            Contains configuration options for the report
        features_imp_list : list
            list of features order by importance
        data_concat : pandas.DataFrame
            Concatanate dataframe of baseline and current datasets
        """

        self.smartdrift = smartdrift
        self.explainer = explainer
        if self.explainer.features_imp is None:
            self.explainer.compute_features_import(force=True)
        self.features_imp_list = self.explainer.features_imp[0].sort_values(ascending=False).index.to_list()  # type: ignore
        self.config = config if config is not None else dict()

        self.data_concat = self._create_data_drift(
            df_current=self.smartdrift.df_current,
            df_baseline=self.smartdrift.df_baseline,
            dataset_names=self.smartdrift.dataset_names,
        )

        if project_info_file is None:
            self.metadata = {"Tip": "You can add custom information here. Check our documentation !"}
        else:
            self.metadata = load_yml(path=project_info_file)

        if "title_story" in self.config.keys():
            self.title_story = self.config["title_story"]
        else:
            self.title_story = "Eurybia report"

    @staticmethod
    def _create_data_drift(
        df_current: Optional[pd.DataFrame], df_baseline: Optional[pd.DataFrame], dataset_names: pd.DataFrame
    ) -> Union[pd.DataFrame, None]:
        """
        Creates a DataFrame that contains dataset used for
        training part and dataset used for production with the column 'data_drift_split'
        allowing to distinguish the values.
        Parameters
        ----------
        df_current : pd.DataFrame, optional
            dataset used for production, dataframe
        df_baseline : pd.DataFrame, optional
            dataset used for traning part, dataframe
        dataset_names : pd.DataFrame
            DataFrame used to specify names to display in report
        Returns
        -------
        pd.DataFrame
            The concatenation of df_baseline and df_current as a dataframe containing df_baseline and df_current values with
            a new 'data_train_test' column allowing to distinguish the values.
        """
        if (df_current is not None and "data_drift_split" in df_current.columns) or (
            df_baseline is not None and "data_drift_split" in df_baseline.columns
        ):
            raise ValueError('"data_drift_split" column must be renamed as it is used in ProjectReport')
        if df_current is None and df_baseline is None:
            return None
        return pd.concat(
            [
                df_current.assign(data_drift_split=dataset_names["df_current"].values[0])
                if df_current is not None
                else None,
                df_baseline.assign(data_drift_split=dataset_names["df_baseline"].values[0])
                if df_baseline is not None
                else None,
            ]
        ).reset_index(drop=True)

    def display_dataset_analysis(self, global_analysis: bool = True, univariate_analysis: bool = True):
        """
        This method performs and displays an exploration of the data given.
        It allows to compare train and test values for each part of the analysis.
        The parameters of the method allow to filter which part to display or not.
        Parameters
        ----------
        global_analysis : bool
            Whether or not to display the global analysis part
        univariate_analysis : bool
            Whether or not to display the univariate analysis part
        target_analysis : bool
            Whether or not to display the target analysis part that plots
            the distribution of the target variable
        multivariate_analysis : bool
            Whether or not to display the multivariate analysis part
        """
        res = {}
        if global_analysis:
            res["global"] = self._display_dataset_analysis_global()

        if univariate_analysis:
            plot_list, labels, table_list = self._perform_and_display_analysis_univariate(
                df=self.data_concat,
                col_splitter="data_drift_split",
                split_values=[
                    self.smartdrift.dataset_names["df_current"].values[0],
                    self.smartdrift.dataset_names["df_baseline"].values[0],
                ],
                names=[
                    self.smartdrift.dataset_names["df_current"].values[0],
                    self.smartdrift.dataset_names["df_baseline"].values[0],
                ],
                group_id="univariate",
            )
            res["univariate"] = (plot_list, labels, table_list)

        return res

    def _display_dataset_analysis_global(self):
        df_stats_global = self._stats_to_table(
            test_stats=perform_global_dataframe_analysis(self.smartdrift.df_current),
            train_stats=perform_global_dataframe_analysis(self.smartdrift.df_baseline),
            names=[
                self.smartdrift.dataset_names["df_current"].values[0],
                self.smartdrift.dataset_names["df_baseline"].values[0],
            ],
        )
        return df_stats_global

    def _perform_and_display_analysis_univariate(
        self,
        df: pd.DataFrame,
        col_splitter: str,
        split_values: list,
        names: list,
        group_id: str,
    ):
        col_types = compute_col_types(df)
        n_splits = df[col_splitter].nunique()
        test_stats_univariate = perform_univariate_dataframe_analysis(
            df.loc[df[col_splitter] == split_values[0]], col_types=col_types
        )
        if n_splits > 1:
            train_stats_univariate = perform_univariate_dataframe_analysis(
                df.loc[df[col_splitter] == split_values[1]], col_types=col_types
            )

        plot_list = []
        labels = []
        table_list = []
        for col in df.drop(col_splitter, axis=1)[self.features_imp_list].columns:
            try:
                fig = self.smartdrift.plot.generate_fig_univariate(df_all=df, col=col, hue=col_splitter)
                plot_list.append(fig)
                labels.append(col)
                df_col_stats = self._stats_to_table(
                    test_stats=test_stats_univariate[col],
                    train_stats=train_stats_univariate[col] if n_splits > 1 else None,
                    names=names,
                )
                table_list.append(df_col_stats)

            except BaseException as e:
                raise Exception(
                    """
                There is a problem with the format of {} column between the two datasets.
                Error:
                """.format(
                        str(col)
                    )
                    + str(e)
                )
        return plot_list, labels, table_list

    @staticmethod
    def _stats_to_table(
        test_stats: dict,
        names: list,
        train_stats: Optional[dict] = None,
    ) -> pd.DataFrame:
        if train_stats is not None:
            return pd.DataFrame({names[1]: pd.Series(train_stats), names[0]: pd.Series(test_stats)})
        else:
            return pd.DataFrame({names[0]: pd.Series(test_stats)})

    def display_model_contribution(self):
        """
        Displays explainability of the model as computed in SmartPlotter object
        """
        multiclass = True if (self.explainer._classes and len(self.explainer._classes) > 2) else False
        c_list = self.explainer._classes if multiclass else [0, 1]  # list just used for multiclass
        plot_list = []
        labels = []
        for index_label, label in enumerate(c_list):  # Iterating over all labels in multiclass case

            for feature in self.features_imp_list:
                fig = self.explainer.plot.contribution_plot(feature, label=label, max_points=200)
                plot_list.append(fig)
                labels.append(feature)
        return plot_list, labels

    def display_data_modeldrift(self):
        """
        Display modeldrift computed metrics when method 'drift' is used
        """
        if self.smartdrift.data_modeldrift is not None:
            plot_list = []
            labels = []
            # If you don't have reference_columns
            if self.smartdrift.data_modeldrift.iloc[:, 1:-1].shape[1] == 0:
                reference_columns = list(self.smartdrift.data_modeldrift.iloc[:, 1:-1].columns)
                metric = "".join(list(self.smartdrift.data_modeldrift.iloc[:, -1:].columns))
                fig = self.smartdrift.plot.generate_modeldrift_data(
                    self.smartdrift.data_modeldrift, metric=metric, reference_columns=reference_columns
                )
                plot_list.append(fig)
            # If you have reference_columns
            else:
                agg_columns = list(self.smartdrift.data_modeldrift.iloc[:, 1:-1].columns)
                for indice, row in self.smartdrift.data_modeldrift[agg_columns].drop_duplicates().iterrows():
                    df = copy.deepcopy(self.smartdrift.data_modeldrift)
                    description = ""
                    for column in agg_columns:
                        df = df[df[column] == row[column]]
                        description += str(row[column]) + ", "

                    reference_columns = list(df.iloc[:, 1:-1].columns)
                    metric = "".join(list(df.iloc[:, -1:].columns))

                    fig = self.smartdrift.plot.generate_modeldrift_data(
                        df, metric=metric, reference_columns=reference_columns
                    )
                    plot_list.append(fig)
                    labels.append(str(description)[:-2])
        return plot_list, labels



"""
Module Unit test of common.py
"""
import os
import unittest
from os import path
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score

from eurybia.report.common import (
    VarType,
    display_value,
    get_callable,
    load_saved_df,
    numeric_is_continuous,
    replace_dict_values,
    series_dtype,
)


class TestCommon(unittest.TestCase):
    """
    Unit test common.py
    """

    def test_series_dtype_1(self):
        """
        Test string series
        """
        serie = pd.Series(["a", "b", "c", "d", "e", np.nan])

        assert series_dtype(serie) == VarType.TYPE_CAT

    def test_series_dtype_2(self):
        """
        Test bool series
        """
        serie = pd.Series([True, True, False, False, False])

        assert series_dtype(serie) == VarType.TYPE_CAT

    def test_series_dtype_3(self):
        """
        Test int and continuous series
        """
        serie = pd.Series(list(range(50)))

        assert series_dtype(serie) == VarType.TYPE_NUM

    def test_series_dtype_4(self):
        """
        Test float and continuous series
        """
        serie = pd.Series(np.linspace(0, 3, 50))

        assert series_dtype(serie) == VarType.TYPE_NUM

    def test_series_dtype_int_5(self):
        """
        Test int and categorical series
        """
        serie = pd.Series([1, 1, 1, 2, 2, 2])

        assert series_dtype(serie) == VarType.TYPE_CAT

    def test_series_dtype_int_6(self):
        """
        Test float and categorical series
        """
        serie = pd.Series([0.2, 0.2, 0.2, 0.6, 0.6, 0.6])

        assert series_dtype(serie) == VarType.TYPE_CAT

    def test_numeric_is_continuous_1(self):
        """
        Test int and continuous series
        """
        serie = pd.Series(list(range(50)))

        assert numeric_is_continuous(serie) is True

    def test_numeric_is_continuous_2(self):
        """
        Test float and continuous series
        """
        serie = pd.Series(np.linspace(0, 1, 100))

        assert numeric_is_continuous(serie) is True

    def test_numeric_is_continuous_3(self):
        """
        Test int and categorical series
        """
        serie = pd.Series([1, 1, 1, 2, 2, 2])

        assert numeric_is_continuous(serie) is False

    def test_numeric_is_continuous_4(self):
        """
        Test float and categorical series
        """
        serie = pd.Series([0.2, 0.2, 0.2, 0.6, 0.6, 0.6])

        assert numeric_is_continuous(serie) is False

    def test_get_callable(self):
        """
        unit test get_callable method
        """
        func = get_callable("sklearn.metrics.accuracy_score")
        y_true = [1, 1, 0, 1, 0]
        y_pred = [1, 1, 1, 0, 0]

        assert accuracy_score(y_true, y_pred) == func(y_true, y_pred)

    def test_display_value_1(self):
        """
        Unit test 1 display_value
        """
        value = 123456.789
        expected_str = "123,456.789"
        assert display_value(value, ",", ".") == expected_str

    def test_display_value_2(self):
        """
        Unit test 2 display_value
        """
        value = 123456.789
        expected_str = "123 456,789"
        assert display_value(value, " ", ",") == expected_str

    def test_display_value_3(self):
        """
        Unit test 3 display_value
        """
        value = 123456.789
        expected_str = "123.456,789"
        assert display_value(value, ".", ",") == expected_str

    def test_replace_dict_values_1(self):
        """
        Unit test 1 replace_dict_values
        """
        dico = {"a": 1234, "b": 0.1234, "c": {"d": 123.456, "e": 1234, "f": {"g": 1234}}}
        expected_d = {"a": "1,234", "b": "0.1234", "c": {"d": "123.456", "e": "1,234", "f": {"g": "1,234"}}}
        res_d = replace_dict_values(dico, display_value, ",", ".")
        assert dico == expected_d
        assert res_d == expected_d

    def test_load_saved_df_1(self):
        """
        Unit test of load_saved_df's method
        """
        file = load_saved_df(path="test")
        assert file is None

    def test_load_saved_df_2(self):
        """
        Unit test of load_saved_df's method
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        y_path = path.join(script_path, "tests/data/test.pkl")
        data_y = pd.DataFrame({"test": [1, 2, 1]})
        data_y.to_pickle(y_path)
        file = load_saved_df(path=y_path)
        assert data_y.equals(file)
        os.remove(y_path)



"""
Statistical test functions
"""
import numpy as np
import pandas as pd
from scipy import stats
import scipy.spatial


def ksmirnov_test(obs_a: np.array, obs_b: np.array) -> dict:
    """
    Returns a dict containing testname, statistic, pvalue of the ks test

    Parameters
    ----------
    obs_a : np.array
        1D array containing the feature values in the first sample
    obs_b : np.array
        1D array containing the feature values ​​in the second sample

    Returns
    -------
    dict :
        3 keys : testname, statistic, pvalue
    """
    test_result = stats.ks_2samp(obs_a, obs_b)
    output = {"testname": "K-Smirnov", "statistic": test_result.statistic, "pvalue": test_result.pvalue}
    return output


def chisq_test(obs_a: np.array, obs_b: np.array) -> dict:
    """
    Returns a dict containing testname, statistic, pvalue of the chisquare test

    Parameters
    ----------
    obs_a : np.array
        1D array containing the feature values in the first sample
    obs_b : np.array
        1D array containing the feature values ​​in the second sample

    Returns
    -------
    dict :
        3 keys : testname, statistic, pvalue
    """
    uniq_a, freq_a = np.unique(obs_a, return_counts=True)
    freq_a_df = pd.DataFrame(freq_a, index=uniq_a, columns=["a"])

    uniq_b, freq_b = np.unique(obs_b, return_counts=True)
    freq_b_df = pd.DataFrame(freq_b, index=uniq_b, columns=["b"])

    freq = pd.concat([freq_a_df, freq_b_df], axis=1).transpose().to_numpy(na_value=0)

    g, p, _, _ = stats.chi2_contingency(freq)

    output = {"testname": "Chi-Square", "statistic": g, "pvalue": p}
    return output



def prob_mass_fun(data, n, range):
    """
    Computing the probability mass function using NumPy’s histogram.
    
    Parameters
    ----------
    data: pandas.Series
        Series to split
    n: int
        The number of equal-width bins in the given range.
    range: tuple
        The lower and upper range of the bins.
    Returns
    -------
    e, Return the bin edges (length= n+1).
    p, Return the frequencies of each interval (length= n).

    """
    h, e = np.histogram(data, n, range)
    p = h / data.shape[0]
    return e, p


def compute_js_divergence(df_1, df_2, n_bins=30):
    """
    Computing the Jensen-Shannon divergence between 2 dataframe

    Parameters
    ----------
    df_1: pandas.Series
        Series to compare
    df_2: pandas.DataFrame
        Series to compare
    n_bins: int
        The number of equal-width bins in the given range
    Returns
    -------
    jensenshannon(p, q) : score between 0 (The two score distributions are identical)
    and 1 (The two score distributions maximally different).
    """
    a = np.concatenate((df_1, df_2), axis=0)
    e, p = prob_mass_fun(df_1, n = n_bins, range = (a.min(), a.max()))
    _, q = prob_mass_fun(df_2, n = e, range = (a.min(), a.max()))

    return scipy.spatial.distance.jensenshannon(p, q)



"""
Common functions used in report
"""
import os
from enum import Enum
from numbers import Number
from typing import Callable, Dict, Optional, Union

import pandas as pd
from pandas.api.types import is_bool_dtype, is_categorical_dtype, is_numeric_dtype, is_string_dtype


class VarType(Enum):
    """
    Helper class to indicate the type of a variable.
    """

    TYPE_CAT = "Categorical"
    TYPE_NUM = "Numeric"
    TYPE_UNSUPPORTED = "Unsupported"

    def __str__(self):
        return str(self.value)


def display_value(value: float, thousands_separator: str = ",", decimal_separator: str = ".") -> str:
    """
    Display a value as a string with specific format.
    Parameters
    ----------
    value : float
        Value to display.
    thousands_separator : str
        The separator used to separate thousands.
    decimal_separator : str
        The separator used to separate decimal values.
    Returns
    -------
    str
    Examples
    --------
    >>> display_value(1255000, thousands_separator=',')
    '1,255,000'
    """
    value_str = f"{value:,}".replace(",", "/thousands/").replace(".", "/decimal/")
    return value_str.replace("/thousands/", thousands_separator).replace("/decimal/", decimal_separator)


def replace_dict_values(obj: Dict, replace_fn: Callable, *args) -> dict:
    """
    Recursively iterates over all values of obj and changes its values using the replace_fn
    Parameters
    ----------
    obj : dict
    replace_fn : callable
    Returns
    -------
    dict
    """
    for k, value in obj.items():
        if isinstance(value, dict):
            obj[k] = replace_dict_values(value, replace_fn)
        elif isinstance(value, Number):
            obj[k] = replace_fn(value, *args)
    return obj


def series_dtype(s: pd.Series) -> VarType:
    """
    Computes the type of a pandas series.
    Parameters
    ----------
    s : pd.Series
        The series for which we wish to determine the type.
    Returns
    -------
    VarType
    """
    if is_bool_dtype(s):
        return VarType.TYPE_CAT
    elif is_string_dtype(s):
        return VarType.TYPE_CAT
    elif is_categorical_dtype(s):
        return VarType.TYPE_CAT
    elif is_numeric_dtype(s):
        if numeric_is_continuous(s):
            return VarType.TYPE_NUM
        else:
            return VarType.TYPE_CAT
    else:
        return VarType.TYPE_UNSUPPORTED


def numeric_is_continuous(s: pd.Series):
    """
    Function that returns True if a
    numeric pandas series is continuous and False if it is categorical.
    Parameters
    ----------
    s : pd.Series
    Returns
    -------
    bool
    """
    # This test could probably be improved
    n_unique = s.nunique()
    return True if n_unique > 15 else False


def get_callable(path: str):
    """
    This function is similar to the _locate function in Hydra library
    Locate an object by name or dotted path, importing as necessary.
    """
    if path == "":
        raise ImportError("Empty path")
    import builtins
    from importlib import import_module

    parts = [part for part in path.split(".") if part]
    module = None
    for number in reversed(range(len(parts))):
        try:
            mod = ".".join(parts[:number])
            module = import_module(mod)
        except Exception as e:
            if number == 0:
                raise ImportError(f"Error loading module '{path}'") from e
            continue
        if module:
            break
    if module:
        obj = module
    else:
        obj = builtins
    for part in parts[number:]:
        mod = mod + "." + part
        if not hasattr(obj, part):
            try:
                import_module(mod)
            except Exception as e:
                raise ImportError(f"Encountered error: `{e}` when loading module '{path}'") from e
        obj = getattr(obj, part)
    if isinstance(obj, type):
        obj_type: type = obj
        return obj_type
    elif callable(obj):
        obj_callable = obj
        return obj_callable
    else:
        # dummy case
        raise ValueError(f"Invalid type ({type(obj)}) found for {path}")


def compute_col_types(df_all: Optional[pd.DataFrame]) -> Dict:
    """
    Computes the type of each column and stores the result in a dict.
    Parameters
    ----------
    df_all : pd.DataFrame, optional
    Returns
    -------
    col_types : dict
        The types of each column
    """
    if df_all is None:
        return {}
    return {col: series_dtype(df_all[col]) for col in df_all.columns}


def load_saved_df(path: str) -> Union[pd.DataFrame, None]:
    """
    Loads a pandas DataFrame that was saved using pd.to_csv method.
    Parameters
    ----------
    path : str
        Path to the dataframe object
    Returns
    -------
    pd.DataFrame or None
    """
    if os.path.exists(path):
        return pd.read_pickle(path)

    else:
        return None



"""
Unit tests for project_report.py
"""
import os
import unittest
from os import path
from pathlib import Path
from unittest.mock import patch

import pandas as pd
import plotly
from category_encoders import OrdinalEncoder
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from eurybia import SmartDrift
from eurybia.report.project_report import DriftReport

expected_attrs = ["smartdrift", "explainer", "config"]

current_path = os.path.dirname(os.path.abspath(__file__))


class TestDriftReport(unittest.TestCase):
    """
    Unit tests for project_report.py
    """

    def setUp(self):
        """
        Use a setUp to init attribute to use in unit test
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, "eurybia/data/titanicdata.csv")
        titan_df = pd.read_csv(titanic_original, index_col=0)
        y = titan_df["Survived"]
        X = titan_df.drop("Survived", axis=1).drop("Name", axis=1)
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(X)
        x_encoded = categ_encoding.transform(X)

        x_train, x_test, y_train, y_test = train_test_split(X, y, train_size=0.75, random_state=1)
        x_train_e, x_test_e, y_train_e, y_test_e = train_test_split(x_encoded, y, train_size=0.75, random_state=1)

        rf = RandomForestClassifier(n_estimators=100, min_samples_leaf=3)
        rf.fit(x_train_e, y_train_e)

        smartdrift = SmartDrift(x_train, x_test, deployed_model=rf, encoding=categ_encoding)
        smartdrift.compile(full_validation=True)

        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smartdrift.add_data_modeldrift(dataset=df_lift, metric="lift", reference_columns=["target", "plage_historique"])
        self.project_info_file = os.path.join(script_path, "eurybia/data/project_info_titanic.yml")
        self.x_train = x_train
        self.x_test = x_test
        self.rf = rf
        self.categ_encoding = categ_encoding
        self.df_list = df_lift
        self.report = DriftReport(
            smartdrift=smartdrift,
            explainer=smartdrift.xpl,
            project_info_file=self.project_info_file,
            config=dict(title_story="Drift Report", title_description="Test drift report"),
        )

    def test_init_1(self):
        """
        Test init() of DriftReport Class
        """
        report = self.report
        for attr in expected_attrs:
            assert hasattr(report, attr)

    def test_init_2(self):
        """
        Test init() of DriftReport Class
        """
        assert isinstance(self.report.features_imp_list, list)

    def test_init_4(self):
        """
        Test init() of DriftReport Class
        """
        assert self.report.title_story == "Drift Report"

    def test_create_data_drift_1(self):
        """
        test _create_data_drift() method
        """
        report = self.report
        df_baseline = pd.DataFrame({"data_drift_split": ["test", "test"]})
        df_current = pd.DataFrame({"data_drift_split": ["test", "test"]})
        dataset_names = pd.DataFrame({"df_current": "Current dataset", "df_baseline": "Historical dataset"}, index=[0])

        with self.assertRaises(ValueError):
            report._create_data_drift(df_current, df_baseline, dataset_names)

    def test_create_data_drift_2(self):
        """
        test _create_data_drift() method
        """
        report = self.report
        df_baseline = None
        df_current = None
        dataset_names = None
        assert report._create_data_drift(df_current, df_baseline, dataset_names) is None

    def test_display_model_contribution(self):
        """
        Test display_model_contribution() method
        """
        self.report.display_model_contribution()

    def test_display_data_modeldrift_1(self):
        """
        Test display_data_modeldrift() method
        """
        fig = self.report.display_data_modeldrift()
        assert isinstance(fig[0][0], plotly.graph_objs._figure.Figure)

    def test_display_data_modeldrift_2(self):
        """
        Test display_data_modeldrift() method
        """
        smartdrift = SmartDrift(self.x_train, self.x_test, deployed_model=self.rf, encoding=self.categ_encoding)
        smartdrift.compile()
        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smartdrift.add_data_modeldrift(dataset=df_lift, metric="lift")
        report = DriftReport(
            smartdrift=smartdrift,
            explainer=smartdrift.xpl,
            config=dict(title_story="Drift Report", title_description="Test drift report"),
        )
        fig = report.display_data_modeldrift()
        assert isinstance(fig[0][0], plotly.graph_objs._figure.Figure)



"""
Unit tests for SmartDrift
"""
import os
import unittest
from os import path
from pathlib import Path
from unittest.mock import Mock, patch

import pandas as pd
import shapash
from category_encoders import OrdinalEncoder
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.model_selection import train_test_split

from eurybia import SmartDrift
from eurybia.style.style_utils import colors_loading, select_palette

TITANIC_PATH = "eurybia/data/titanicdata.csv"
TITANIC_ORIGINAL_PATH = "eurybia/data/titanic_original.csv"
TITANIC_ALTERED_PATH = "eurybia/data/titanic_altered.csv"
PROJECT_INFO_PATH = "eurybia/data/project_info_titanic.yml"
REPORT_PATH = "./test_report.html"


class TestSmartDrift(unittest.TestCase):
    """
    Unit test on SmartDrift class
    """

    def setUp(self):
        """SetUp"""
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_csv = path.join(script_path, TITANIC_PATH)
        titanic_df = pd.read_csv(titanic_csv)
        del titanic_df["Name"]
        titanic_df = titanic_df.drop("PassengerId", axis=1)
        titanic_df_1, titanic_df_2 = train_test_split(titanic_df, test_size=0.5, random_state=42)
        titanic_df_1 = titanic_df_1.drop("Survived", axis=1)
        titanic_df_2 = titanic_df_2.drop("Survived", axis=1)
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        y = titanic_df["Survived"]
        X = titanic_df.drop("Survived", axis=1)
        categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(X)
        X = categ_encoding.transform(X)
        x_train, x_test, y_train, y_test = train_test_split(X, y, train_size=0.75, random_state=1)
        rf = RandomForestClassifier(n_estimators=100, min_samples_leaf=3)
        rf.fit(x_train, y_train)
        self.titanic_df_1 = titanic_df_1
        self.titanic_df_2 = titanic_df_2
        self.rf = rf
        self.categ_encoding = categ_encoding
        self.script_path = script_path
        self.X = X

    def test_init_1(self):
        """
        test init 1 SmartDrift
        """
        smart_drift = SmartDrift()
        assert hasattr(smart_drift, "df_baseline")

    def test_compile_nooptions(self):
        """
        Test compile()
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile()
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_fullvalid(self):
        """
        test compile() with full validation
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_nosampling(self):
        """
        test compile() without no sampling
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True, sampling=False)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_samplingsize(self):
        """
        test compile() with a fixed sample size
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True, sample_size=100)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_model_encoder(self):
        """
        test compile() with a model and an encoder specified
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile()
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_generate_report_fullvalid(self):
        """
        test generate_report() with fullvalidation option specified to True
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(full_validation=True)
        infofile = path.join(self.script_path, PROJECT_INFO_PATH)
        smart_drift.generate_report(output_file=REPORT_PATH, project_info_file=infofile)

        assert path.isfile(REPORT_PATH)
        os.remove(REPORT_PATH)

    def test_generate_report_nofullvalid(self):
        """
        test generate_report() with no full validation
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile()
        infofile = path.join(self.script_path, PROJECT_INFO_PATH)
        smart_drift.generate_report(output_file=REPORT_PATH, project_info_file=infofile)
        assert path.isfile(REPORT_PATH)
        os.remove(REPORT_PATH)

    def test_add_data_modeldrift_1(self):
        """
        test add_data_modeldrift method
        """
        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "performance": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(ignore_cols=["PassengerId"])
        assert isinstance(smart_drift.data_modeldrift, type(None))
        smart_drift.add_data_modeldrift(dataset=df_lift)
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)

    def test_add_data_modeldrift_2(self):
        """
        test add_data_modeldrift method
        """
        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(ignore_cols=["PassengerId"])
        assert isinstance(smart_drift.data_modeldrift, type(None))
        with self.assertRaises(Exception):
            smart_drift.add_data_modeldrift(dataset=df_lift)
        smart_drift.add_data_modeldrift(dataset=df_lift, metric="lift")
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)
        smart_drift.add_data_modeldrift(
            dataset=df_lift, metric="lift", reference_columns=["target", "plage_historique"]
        )
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)

    def test_analyse_consistency_nofullvalidation_1(self):
        """
        Test _analyze_consistency() method with full validation defined to False
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"]
        )

        expected_pb_cols = {"New columns": [], "Removed columns": ["col_sup"], "Type errors": []}

        assert isinstance(pb_cols, dict)
        assert all(key in ("New columns", "Removed columns", "Type errors") for key in pb_cols.keys())
        assert isinstance(err_mods, dict)
        assert pb_cols == expected_pb_cols
        assert err_mods == dict()

    def test_analyse_consistency_nofullvalidation_2(self):
        """
        Test _analyze_consistency() method with full validation defined to False
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"]
        )

        with self.assertRaises(TypeError):
            smart_drift.df_current = dict()
            smart_drift._analyze_consistency(full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"])

    def test_analyse_consistency_fullvalidation(self):
        """
        Test _analyze_consistency() method with full validation defined to True
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        df_altered = df_altered.replace("male", "Male")
        df_original["col_sup"] = "0"
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=True, ignore_cols=["Title", "Name", "PassengerId"]
        )

        expected_err_mods = {"Sex": {"New distinct values": ["Male"], "Removed distinct values": ["male"]}}

        expected_pb_cols = {"New columns": [], "Removed columns": [], "Type errors": ["col_sup"]}

        assert isinstance(pb_cols, dict)
        assert all(key in ("New columns", "Removed columns", "Type errors") for key in pb_cols.keys())
        assert isinstance(err_mods, dict)
        assert pb_cols == expected_pb_cols
        assert err_mods == expected_err_mods

    def test_predict_1(self):
        """
        test _predict() method
        """

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict()
        assert df_predict is None

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf)

        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)
        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_predict_2(self):
        """
        test _predict() method
        """

        varcat = ["Pclass", "Sex", "Embarked", "Title"]

        false_categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(
            self.X[["Title", "Pclass", "Sex", "Embarked"]]
        )
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)

        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_predict_3(self):
        """
        test _predict() method
        """
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        false_categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(
            self.X[["Title", "Pclass", "Sex", "Embarked"]]
        )
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)

        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)
        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_feature_importance(self):
        """
        Test on _feature_importance()'s method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile()

        feature_imp = smart_drift._feature_importance()
        assert feature_imp is None

        feature_imp = smart_drift._feature_importance(deployed_model=self.rf)
        assert isinstance(feature_imp, pd.DataFrame)
        assert all(column in ("feature", "deployed_model", "datadrift_classifier") for column in feature_imp.columns)
        assert all(col_type in ("float64", "object") for col_type in feature_imp.dtypes)

    def test_histo_datadrift_metric_1(self):
        """
        Test on _histo_datadrift_metric() method
        """
        SD = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_auc = SD._histo_datadrift_metric()
        assert df_auc == None

        SD = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        SD.compile(datadrift_file="tests/data/AUC_Histo.csv")
        df_auc = SD._histo_datadrift_metric()
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_2(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(
                datadrift_file="tests/data/AUC_Histo.csv", predict_test=smart_drift.model.predict(self.titanic_df_2)
            )

        df_auc = smart_drift._histo_datadrift_metric()
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_3(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(date_compile_auc="01-12-2020")

        datadrift_file = path.join(self.script_path, "tests/data/AUC_Histo.csv")
        df_auc = smart_drift._histo_datadrift_metric(date_compile_auc="01/12/2020", datadrift_file=datadrift_file)
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_4(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(date_compile_auc="01-12-2020")

        datadrift_file = path.join(self.script_path, "tests/data/AUC_Histo.csv")
        df_auc = smart_drift._histo_datadrift_metric(date_compile_auc="01/12/2020", datadrift_file=datadrift_file)
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc", "JS_predict") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_save_load(self):
        """
        Test on save() and load() methods : checks that the attributes meant to be saved and loaded are saved and loaded
        """
        temp_pkl_path = "tests/data/test_pkl.pkl"
        sd = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        sd.compile(datadrift_file="tests/data/AUC_Histo.csv")
        sd.save(temp_pkl_path)
        sd2 = SmartDrift.load(temp_pkl_path)
        assert isinstance(sd2, SmartDrift)
        pd.testing.assert_frame_equal(sd.df_baseline, sd2.df_baseline)
        assert len(sd.df_baseline) == len(sd2.df_baseline)
        self.assertCountEqual(sd.df_current.columns, sd2.df_current.columns)
        assert len(sd.df_current) == len(sd2.df_current)
        assert sd.feature_importance == sd2.feature_importance
        self.assertCountEqual(sd.ignore_cols, sd2.ignore_cols)
        self.assertCountEqual(sd.pb_cols, sd2.pb_cols)
        self.assertCountEqual(sd.err_mods, sd2.err_mods)
        assert sd.auc == sd2.auc
        pd.testing.assert_frame_equal(sd.historical_auc, sd2.historical_auc)
        pd.testing.assert_frame_equal(sd.datadrift_stat_test, sd2.datadrift_stat_test)
        pd.testing.assert_frame_equal(sd._df_concat, sd2._df_concat)
        assert sd._datadrift_target == sd2._datadrift_target
        assert sd.deployed_model == sd2.deployed_model
        assert sd.encoding == sd2.encoding
        assert sd.palette_name == sd2.palette_name
        assert sd.colors_dict == sd2.colors_dict
        os.remove(temp_pkl_path)

    def test_define_style(self):
        """
        test define_style() method : checks that the frontend style defined in colors.json is written into the SmartDrift plot attribute
        """
        sd = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        sd.compile(datadrift_file="tests/data/AUC_Histo.csv")
        colors_dict = select_palette(colors_loading(), "eurybia")
        with self.assertRaises(ValueError):
            sd.define_style()
        sd.define_style(palette_name="eurybia", colors_dict=colors_dict)
        assert sd.colors_dict == colors_dict
        assert sd.plot._style_dict["univariate_cat_bar"] == colors_dict["univariate_cat_bar"]
        assert sd.plot._style_dict["univariate_cont_bar"] == colors_dict["univariate_cont_bar"]
        assert sd.plot._style_dict["datadrift_historical"] == colors_dict["datadrift_historical"]
        assert sd.plot._style_dict["scatter_plot"] == colors_dict["scatter_plot"]
        assert sd.plot._style_dict["scatter_line"] == colors_dict["scatter_line"]
        assert sd.plot._style_dict["featimportance_colorscale"] == colors_dict["featimportance_colorscale"]
        assert sd.plot._style_dict["contrib_colorscale"] == colors_dict["contrib_colorscale"]
        # not testing the shapash.explainer.smart_explainer



"""Top-level package."""

__author__ = """Thomas Bouche, Johann Martin, Nicolas Roux"""
__email__ = "thomas.bouche@maif.fr"

from eurybia.core.smartdrift import SmartDrift

from .__version__ import __version__

__all__ = ["SmartDrift"]



#!/usr/bin/env python

"""The setup script."""
import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", encoding="utf8") as readme_file:
    long_description = readme_file.read()

# Load the package's __version__.py module as a dictionary.
version_d: dict = {}
with open(os.path.join(here, "eurybia", "__version__.py")) as f:
    exec(f.read(), version_d)

requirements = [
    "catboost>=0.22",
    "datapane==0.14.0",
    "ipywidgets>=7.4.2",
    "jinja2>=2.11.0,<3.1.0",
    "shapash>=2.0.0",
    "seaborn<=0.11.1",
    "scipy>=1.4.0",
    "jupyter",
]


setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    name="eurybia",  # Replace with your own username
    version=version_d["__version__"],
    python_requires=">3.6, < 3.10",
    url="https://github.com/MAIF/eurybia",
    author="Nicolas Roux, Johann Martin, Thomas Bouché",
    author_email="thomas.bouche@maif.fr",
    description="Eurybia monitor model drift over time and securize model deployment with data validation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    license="Apache Software License 2.0",
    keywords="eurybia",
    package_dir={
        "eurybia": "eurybia",
        "eurybia.data": "eurybia/data",
        "eurybia.core": "eurybia/core",
        "eurybia.report": "eurybia/report",
        "eurybia.style": "eurybia/style",
        "eurybia.utils": "eurybia/utils",
    },
    packages=["eurybia", "eurybia.data", "eurybia.core", "eurybia.report", "eurybia.style", "eurybia.utils"],
    data_files=[
        ("data", ["eurybia/data/house_prices_dataset.csv"]),
        ("data", ["eurybia/data/house_prices_labels.json"]),
        ("data", ["eurybia/data/titanicdata.csv"]),
        ("data", ["eurybia/data/project_info_car_accident.yml"]),
        ("data", ["eurybia/data/project_info_house_price.yml"]),
        ("data", ["eurybia/data/project_info_titanic.yml"]),
        ("data", ["eurybia/data/titanic_altered.csv"]),
        ("data", ["eurybia/data/titanic_original.csv"]),
        ("data", ["eurybia/data/US_Accidents_extract.csv"]),
        ("style", ["eurybia/style/colors.json"]),
    ],
    include_package_data=True,
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    zip_safe=False,
)



"""
- SmartDrift module
"""
import copy
import datetime
import io
import logging
import pickle
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Dict, Text

import catboost
import pandas as pd
from shapash.explainer.smart_explainer import SmartExplainer
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from eurybia.core.smartplotter import SmartPlotter
from eurybia.style.style_utils import colors_loading, select_palette
from eurybia.utils.io import load_pickle, save_pickle
from eurybia.utils.model_drift import catboost_hyperparameter_init, catboost_hyperparameter_type
from eurybia.utils.statistical_tests import chisq_test, compute_js_divergence, ksmirnov_test
from eurybia.utils.utils import base_100

logging.getLogger("papermill").setLevel(logging.WARNING)
logging.getLogger("blib2to3").setLevel(logging.WARNING)


class SmartDrift:
    """
    The SmartDrift class is the main object to compute drift in the Eurybia library
    It allows to calculate data drift between 2 datasets using a data drift classification model

    Attributes
    ----------
    df_current: pandas.DataFrame
        current (or production) dataset which is compared to df_baseline
    df_baseline: pandas.DataFrame
        baseline (or learning) dataset which is compared to df_current
    datadrift_classifier: model object
        model used for binary classification of data drift
    xpl: Shapash object
        object used to compute explainability on datadrift_classifier
    df_predict: pandas.DataFrame
        computed score on both datasets if a deployed_model is specified
    feature_importance: pandas.DataFrame
        feature importance of datadrift_classifier and feature importance of production model if exist
    pb_cols: dict
        Dictionnary that references columns differences between df_current and df_baseline
    err_mods: dict
        Dictionnary that references modalities differences in columns between df_current and df_baseline
    auc: int
        Value auc of model drift
    historical_auc: pandas.DataFrame
        Dataframe that contains auc history of datadrift_classifier over time
    data_modeldrift: pandas.DataFrame
        Dataframe that contains performance history of deployed_model
    ignore_cols: list
        list of feature to ignore in compute
    dataset_names : dict, (Optional)
        Dictionnary used to specify dataset names to display in report.
    _df_concat : pandas.DataFrame
        Dataframe that's composed of both df_baseline and df_current concatenated
    plot : eurybia.core.smartplotter.SmartPlotter
        Instance of an Eurybia SmartPlotter class. It's used for graph displaying purpose.
    deployed_model: model object, optional
            model in production used to put in perspective drift and to predict
    encoding: preprocessing object, optional (default: None)
            Preprocessing used before the training step
    datadrift_stat_test : dict
        Datadrift statistical tests for each feature.
        Each test identifies whether the feature has drifted.
        There are 2 types of test implemented depending on the type of feature:
        - Chi-square for discrete variables - ref:
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chi2_contingency.html)
        - Kolmogorov-Smirnov for continuous variables - ref:
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html)
        This datadrift_stat_test attribute specifies for each feature the test performed,
        the statistic the test and the p value
    palette_name : str (default: 'eurybia')
        Name of the palette used for the colors of the report (refer to style folder).
    colors_dict: dict
            Dict of the colors used in the different plots
    datadrift_file : str, optional
            Name of the csv file that contains the performance history of data drift
            If no datadrift file is given, the drift will not be logged
    js_divergence : float
        Jensen-Shannon divergence of probability distributions - ref:
        (https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence)

    How to declare a new SmartDrift object?

    Example
    --------
    >>> SD = Smartdrift(df_current=df_production, df_baseline=df_learning)

    """

    @classmethod
    def load(cls, path):
        """
        The load() class method allows Eurybia users to use a pickled SmartDrift.

        Parameters
        ----------
        path : str
            File path of the pickle file.
        Returns
        ----------
        SmartDrift
            SmartDrift instance loaded with the pickle given as "path"
        Example
        --------
        >>> from eurybia import SmartDrift
        >>> SmartDrift.load('path_to_pkl/smardrift.pkl')
        """
        dict_to_load = load_pickle(path)
        sd = cls()
        if isinstance(dict_to_load, dict):
            for attr, val in dict_to_load.items():
                if isinstance(val, io.BytesIO):
                    setattr(sd, attr, pickle.load(val.seek(0)))
                elif attr == "xpl":
                    xpl = SmartExplainer(model=val["model"])
                    xpl.__dict__.update(val)
                    setattr(sd, attr, xpl)
                else:
                    setattr(sd, attr, val)
        else:
            raise ValueError("pickle file must contain dictionary")
        return sd

    def __init__(
        self,
        df_current=None,
        df_baseline=None,
        dataset_names={"df_current": "Current dataset", "df_baseline": "Baseline dataset"},
        deployed_model=None,
        encoding=None,
        palette_name="eurybia",
        colors_dict=None,
    ):
        """
        Parameters
        ----------
        df_current: pandas.DataFrame
            current (or production) dataset which is compared to df_baseline
        df_baseline: pandas.DataFrame
            baseline (or learning) dataset which is compared to df_current
        dataset_names : dict, (Optional)
            Dictionnary used to specify dataset names to display in report.
        deployed_model: model object, optional
                model in production used to put in perspective drift and to predict
        encoding: preprocessing object, optional (default: None)
                Preprocessing used before the training step
        palette_name : str (default: 'eurybia')
            Name of the palette used for the colors of the report (refer to style folder).
        colors_dict: dict
                Dict of the colors used in the different plots

        How to declare a new SmartDrift object ?

        Example
        --------

        >>> SD = Smartdrift(df_current=df_production, df_baseline=df_learning)
        """
        self.df_current = df_current
        self.df_baseline = df_baseline
        self.xpl = None
        self.df_predict = None
        self.feature_importance = None
        self.pb_cols, self.err_mods = None, None
        self.auc = None
        self.js_divergence = None
        self.historical_auc = None
        self.data_modeldrift = None
        self.ignore_cols = list()
        self.datadrift_stat_test = None
        if "df_current" not in dataset_names.keys() or "df_baseline" not in dataset_names.keys():
            raise ValueError("dataset_names must be a dictionnary with keys 'df_current' and 'df_baseline'")
        self.dataset_names = pd.DataFrame(dataset_names, index=[0])
        self._df_concat = None
        self._datadrift_target = "target"
        self.plot = SmartPlotter(self)
        self.deployed_model = deployed_model
        self.encoding = encoding
        self.palette_name = palette_name
        self.colors_dict = copy.deepcopy(select_palette(colors_loading(), self.palette_name))
        if colors_dict is not None:
            self.colors_dict.update(colors_dict)
        self.plot.define_style_attributes(colors_dict=self.colors_dict)
        self.datadrift_file = None

    def compile(
        self,
        full_validation=False,
        ignore_cols: list = list(),
        sampling=True,
        sample_size=100000,
        datadrift_file=None,
        date_compile_auc=None,
        hyperparameter: Dict = catboost_hyperparameter_init.copy(),
        attr_importance="feature_importances_",
    ):
        r"""
        The compile method is the first step to compute data drift.
        It allows to calculate data drift between 2 datasets using a data drift classification model.
        Most of the parameters are optional but helps to adapt the data drift calculation if necessary.
        This step can last a few moments with large datasets.

        Parameters
        ----------
        full_validation: bool, optional (default: False)
            If True, analyze consistency on modalities between columns
        ignore_cols: list, optional
            list of feature to ignore in compute
        sampling: bool, optional
            If True, applies the sampling
        sample_size: int, optional
            the size of the sample to build
        date_compile_auc: str (optional)
            format dd/mm/yyyy use for specify date of compute drift, useful when compute few time drift for different time at the same moment
        hyperparameter: dict, optional
            if user want to modify catboost hyperparameter
        attr_importance: string, optional (default: "feature_importances\_")
            Attribute "feature_importance" of the deployed_model
        datadrift_file : str, optional
            Name of the csv file that contains the performance history of data drift. If no datadrift file is given, the drift will not be logged

        Examples
        --------
        >>> SD.compile()

        """
        if datadrift_file is not None:
            self.datadrift_file = datadrift_file
        if hyperparameter is not None:
            for key, value in catboost_hyperparameter_init.items():
                catboost_hyperparameter_init[key] = (
                    hyperparameter[key]
                    if key in hyperparameter and str(type(hyperparameter[key])) in catboost_hyperparameter_type[key]
                    else value
                )
        hyperparameter = catboost_hyperparameter_init.copy()
        if sample_size is not None:
            self.df_baseline = self._sampling(sampling, sample_size, self.df_baseline)
            self.df_current = self._sampling(sampling, sample_size, self.df_current)
        # Consistency analysis
        pb_cols, err_mods = self._analyze_consistency(full_validation, ignore_cols)

        # Adding results to ignored columns
        ignore_cols = list(set(ignore_cols + [item for sublist in [pb_cols[sl] for sl in pb_cols] for item in sublist]))
        if len(ignore_cols) != 0:
            self.df_baseline = self.df_baseline[[c for c in self.df_baseline.columns if c not in ignore_cols]]
            self.df_current = self.df_current[[c for c in self.df_current.columns if c not in ignore_cols]]

        df_concat = (
            pd.concat([self.df_current, self.df_baseline], keys=[1, 0])
            .reset_index()
            .rename(columns={"level_0": self._datadrift_target})
        )
        df_concat.drop(df_concat.columns[1], axis=1, inplace=True)
        varz = [c for c in df_concat.columns if c not in [self._datadrift_target] and c not in ignore_cols]
        dtypes = df_concat[varz].dtypes.map(str)
        cat_features = list(dtypes[dtypes.isin(["object"])].index)
        df_concat[cat_features] = df_concat[cat_features].fillna("NA")
        self._df_concat = df_concat

        train, test = train_test_split(df_concat[varz + [self._datadrift_target]], test_size=0.25, random_state=42)

        i = 0
        indice_cat = []
        for var_x in df_concat[varz]:
            if var_x in cat_features:
                indice_cat.append(i)
            i = i + 1
        train_pool_cat = catboost.Pool(data=train[varz], label=train["target"].astype(int), cat_features=indice_cat)
        test_pool_cat = catboost.Pool(data=test[varz], label=test["target"].astype(int), cat_features=indice_cat)
        datadrift_classifier = catboost.CatBoostClassifier(
            max_depth=hyperparameter["max_depth"],
            l2_leaf_reg=hyperparameter["l2_leaf_reg"],
            learning_rate=hyperparameter["learning_rate"],
            iterations=hyperparameter["iterations"],
            use_best_model=hyperparameter["use_best_model"],
            custom_loss=hyperparameter["custom_loss"],
            loss_function=hyperparameter["loss_function"],
            eval_metric=hyperparameter["eval_metric"],
            task_type="CPU",
        )

        datadrift_classifier = datadrift_classifier.fit(train_pool_cat, eval_set=test_pool_cat, silent=True)

        xpl = SmartExplainer(
            label_dict={0: self.dataset_names["df_baseline"].values[0], 1: self.dataset_names["df_current"].values[0]},
            model=datadrift_classifier,
        )

        x_test = test[varz]
        y_test = test[self._datadrift_target]

        xpl.compile(x=x_test)
        xpl.compute_features_import(force=True)

        self.xpl = xpl
        self.xpl.define_style(colors_dict=self.colors_dict)
        self.datadrift_classifier = datadrift_classifier
        self.df_predict = self._predict(deployed_model=self.deployed_model, encoding=self.encoding)
        self.auc = roc_auc_score(y_test, datadrift_classifier.predict(x_test))
        self.feature_importance = self._feature_importance(
            deployed_model=self.deployed_model, attr_importance=attr_importance
        )
        self.plot.feature_importance = self.feature_importance
        self.pb_cols, self.err_mods = pb_cols, err_mods
        if self.deployed_model is not None:
            self.js_divergence = compute_js_divergence(
                self.df_predict.loc[lambda df: df["dataset"] == "Baseline dataset", :]["Score"].values,
                self.df_predict.loc[lambda df: df["dataset"] == "Current dataset", :]["Score"].values,
                n_bins=20,
            )

        self.historical_auc = self._histo_datadrift_metric(
            datadrift_file=self.datadrift_file, date_compile_auc=date_compile_auc
        )
        self.data_modeldrift = None
        self.ignore_cols = ignore_cols
        self.datadrift_stat_test = self._compute_datadrift_stat_test()

    def generate_report(
        self, output_file, project_info_file=None, title_story="Drift Report", title_description="", working_dir=None
    ):
        """
        This method will generate an HTML report containing different information about the project.
        It allows the information compiled to be rendered.
        It can be associated with a project info yml file on which can figure different information about the project.

        Parameters
        ----------
        output_file : str
            Path to the HTML file to write
        project_info_file : str
            Path to the file used to display some information about the project in the report
        title_story : str, optional
            Report title
        title_description : str, optional
            Report title description (as written just below the title)
        working_dir : str, optional
            Working directory in which will be generated the notebook used to create the report and where the objects used to execute it will
            be saved. This parameter can be usefull if one wants to create its own custom report and debug the notebook used to generate
            the html report. If None, a temporary directory will be used

        Examples
        --------
        >>> SD.generate_report(
                output_file='report.html',
                project_info_file='project_info.yml',
                title_story="Drift project report",
                title_description="This document is a drift report of the score in production"
            )
        """
        from eurybia.report.generation import execute_report

        rm_working_dir = False
        if not working_dir:
            working_dir = tempfile.mkdtemp()
            rm_working_dir = True

        try:
            execute_report(
                project_info_file=project_info_file,
                explainer=self.xpl,
                smartdrift=self,
                config=dict(title_story=title_story, title_description=title_description),
                output_file=output_file,
            )
        finally:
            if rm_working_dir:
                shutil.rmtree(working_dir)

    def _analyze_consistency(self, full_validation=False, ignore_cols: list = list()):
        """
        method to analyse consistency between the 2 datasets, in terms of columns and modalities

        Parameters
        ----------
        full_validation : bool, optional (default: False)
            If True, analyze consistency on modalities between columns
        ignore_cols: list, optional
            list of feature to ignore in compute
        """
        logging.basicConfig(
            level=logging.INFO, format="%(asctime)s %(levelname)s %(module)s: %(message)s", datefmt="%y/%m/%d %H:%M:%S"
        )

        if not isinstance(self.df_current, pd.DataFrame) or not isinstance(self.df_baseline, pd.DataFrame):
            raise TypeError("df_current and df_baseline should be Pandas dataframes.")
        if len(ignore_cols) > 0:
            logging.info(
                """The following variables are set in ignore_cols and
                        will not be analyzed: \n %s""",
                ignore_cols,
            )
        # Features
        new_cols = [c for c in self.df_baseline.columns if c not in self.df_current.columns]
        removed_cols = [c for c in self.df_current.columns if c not in self.df_baseline.columns]
        if len(new_cols) > 0:
            logging.info(
                """
            The following variables are no longer available in the
            current dataset and will not be analyzed: \n %s""",
                new_cols,
            )
        if len(removed_cols) > 0:
            logging.info(
                """
            The following variables are only available in the
            current dataset and will not be analyzed: \n %s""",
                removed_cols,
            )
        common_cols = [c for c in self.df_current.columns if c in self.df_baseline.columns]
        # dtypes
        err_dtypes = [
            c for c in common_cols if self.df_baseline.dtypes.map(str)[c] != self.df_current.dtypes.map(str)[c]
        ]
        if len(err_dtypes) > 0:
            logging.info(
                """
            The following variables have mismatching dtypes
             and will not be analyzed: \n %s""",
                err_dtypes,
            )
        # Feature values
        err_mods: Dict[Text, Dict] = {}
        variables_mm_mods = []
        if full_validation is True:
            invalid_cols = ignore_cols + new_cols + removed_cols + err_dtypes
            for column in self.df_baseline.columns:
                if column not in invalid_cols and self.df_baseline.dtypes.map(str)[column] == "object":
                    uniques_histo = pd.unique(self.df_baseline[column])
                    uniques_current = pd.unique(self.df_current[column])
                    new_mods = [mod for mod in uniques_current if mod not in uniques_histo]
                    removed_mods = [mod for mod in uniques_histo if mod not in uniques_current]
                    if len(new_mods) > 0 or len(removed_mods) > 0:
                        err_mods[column] = {}
                        err_mods[column]["New distinct values"] = new_mods
                        err_mods[column]["Removed distinct values"] = removed_mods
                        logging.info(
                            """The variable %s
                                     has mismatching possible values: \n
                                     %s %s""",
                            column,
                            new_mods,
                            removed_mods,
                        )
                        variables_mm_mods.append(column)
        return ({"New columns": new_cols, "Removed columns": removed_cols, "Type errors": err_dtypes}, err_mods)

    def _predict(self, deployed_model=None, encoding=None):
        """
        Create an attributes df_predict with the computed score on both datasets

        Parameters
        ----------
        deployed_model : model object, optional (default: None)
            model in production used to put in perspective drift and to predict
        encoding : preprocessing object, optional (default: None)
            Preprocessing used before the training step
        Returns
        -------
        pandas.DataFrame, None
            DataFrame with predicted score for both datasets
        """
        if deployed_model is None:
            return None
        if not hasattr(deployed_model, "predict_proba") and not hasattr(deployed_model, "predict"):
            raise Exception("deployed_model need to have predict or predict_proba method")
        df_baseline = self.df_baseline
        df_current = self.df_current
        if encoding is not None:
            try:
                df_baseline = encoding.transform(df_baseline)
                df_current = encoding.transform(df_current)
            except BaseException as error:
                raise Exception(
                    """
                    Encoding specified can't be applied directly on df_current/df_baseline
                    - Error :
                                    """
                    + str(error)
                )
        if hasattr(deployed_model, "predict_proba"):
            try:
                df_baseline_pred = pd.DataFrame(deployed_model.predict_proba(df_baseline)[:, 1], columns=["Score"])
                df_current_pred = pd.DataFrame(deployed_model.predict_proba(df_current)[:, 1], columns=["Score"])
            except BaseException as error:
                raise Exception(
                    """
                    Encoding specified or deployed_model used can't be applied directly on df_current/df_baseline
                    - Error :
                                    """
                    + str(error)
                )
        else:
            try:
                df_baseline_pred = pd.DataFrame(deployed_model.predict(df_baseline), columns=["Score"])
                df_current_pred = pd.DataFrame(deployed_model.predict(df_current), columns=["Score"])
            except BaseException as error:
                raise Exception(
                    """
                    Encoding specified or deployed_model used can't be applied directly on df_current/df_baseline
                    - Error :
                                    """
                    + str(error)
                )
        return pd.concat(
            [
                df_baseline_pred.assign(dataset=self.dataset_names["df_baseline"].values[0]),
                df_current_pred.assign(dataset=self.dataset_names["df_current"].values[0]),
            ]
        ).reset_index(drop=True)

    def _feature_importance(self, deployed_model=None, attr_importance="feature_importances_"):
        """
        Create an attributes feature_importance with the computed score on both datasets

        Parameters
        ----------
        deployed_model : model object, optional (default: None)
            model in production used to put in perspective drift and to predict
        attr_importance : string, optional (default: "feature_importances_")
            Attribute "feature_importance" of the deployed_model
        Returns
        -------
        pandas.DataFrame, None
            DataFrame with feature importance from production model
            and drift model.
        """
        if deployed_model is None:
            return None
        try:
            array_importance = getattr(deployed_model, attr_importance)
        except BaseException as error:
            raise Exception(
                """
            deployed_model used can't allow to get features importance on df_baseline
            - Error :
                            """
                + str(error)
            )
        feature_importance_drift = pd.DataFrame(
            self.xpl.features_imp[0].values, index=self.xpl.features_imp[0].index, columns=["datadrift_classifier"]
        )
        var_baseline = [c for c in self.df_baseline.columns if c not in ["target"]]
        if len(array_importance) != len(var_baseline):
            raise ValueError(
                """
            Number of features in df_baseline doesn't match feature importance's shape returned by deployed model.
            """
            )
        feature_importance_model_prod = pd.DataFrame(
            array_importance, index=self.df_baseline[var_baseline].columns, columns=["deployed_model"]
        )
        feature_importance = feature_importance_model_prod.merge(
            feature_importance_drift, how="left", left_index=True, right_index=True
        ).reset_index()
        feature_importance = feature_importance.rename(columns={"index": "feature"})
        feature_importance["deployed_model"] = base_100(feature_importance["deployed_model"])
        return feature_importance

    def _sampling(self, sampling, sample_size, dataset):
        """
        Return a sampling from the original dataframe

        Parameters
        ----------
        sampling : bool
            If True, applies the sampling
        sample_size : int
            the size of the sample to build
        df : pd.DataFrame
            The Dataframe to apply sampling
        Returns
        -------
        pandas.DataFrame
            a sample of the original DataFrame or the original DataFrame
        """
        if sampling:
            if dataset.shape[0] > sample_size:
                return dataset.sample(sample_size)
            else:
                return dataset
        else:
            return dataset

    def _histo_datadrift_metric(self, datadrift_file=None, date_compile_auc=None):
        """
        Method which computes datadrift metrics (AUC, and Jensen Shannon prediction divergence if the deployed_model is filled in)
        and append it into a dataframe that will be exported during the generate_report method

        Parameters
        ----------
        datadrift_file : str, (optional)
        date_compile_auc: str (optional)
            format dd/mm/yyyy use for specify date of compute drift, useful when compute few time drift for different time at the same moment
        Returns
        -------
        pandas.DataFrame or None
        Dataframe with dates, AUC and Jensen Shannon prediction divergence computed at this date
        """
        logging.basicConfig(
            level=logging.INFO, format="%(asctime)s %(levelname)s %(module)s: %(message)s", datefmt="%y/%m/%d %H:%M:%S"
        )
        if self.datadrift_file is None and date_compile_auc is None:
            return None
        elif self.datadrift_file is None and date_compile_auc is not None:
            self.datadrift_file = "historical_AUC.csv"
            return None
        else:
            if date_compile_auc:
                try:
                    datetime.datetime.strptime(date_compile_auc, "%d/%m/%Y")
                    # FIXME find exact exception type and use it instead of "Exception"
                except Exception:
                    raise Exception("The argument date must have the format '%d/%m/%Y'")
            else:
                date_compile_auc = (datetime.datetime.today().date()).strftime("%d/%m/%Y")
            logging.info(
                """
            The computed AUC on the X_test used to build datadrift_classifier is equal to: {}
                        """.format(
                    str(self.auc)
                )
            )
            if self.deployed_model is not None:
                df_auc = pd.DataFrame(
                    {"date": [date_compile_auc], "auc": [self.auc], "JS_predict": [self.js_divergence]}
                )
            else:
                df_auc = pd.DataFrame({"date": [date_compile_auc], "auc": [self.auc]})

            if self.datadrift_file is not None:
                if Path(self.datadrift_file).is_file() and self.datadrift_file.endswith(".csv"):
                    histo_auc = pd.read_csv(self.datadrift_file).reset_index(drop=True)
                    if self.deployed_model is not None:
                        if not (
                            any(histo_auc.columns.isin(["date"]))
                            and any(histo_auc.columns.isin(["auc"]))
                            and any(histo_auc.columns.isin(["JS_predict"]))
                        ):
                            raise Exception("The csv data must have columns 'date', 'auc' and 'JS_predict'")
                        df_auc = pd.concat([histo_auc[["date", "auc", "JS_predict"]], df_auc]).reset_index(drop=True)

                    else:
                        if not (any(histo_auc.columns.isin(["date"])) and any(histo_auc.columns.isin(["auc"]))):
                            raise Exception("The csv data must have columns 'date' and 'auc'")
                        df_auc = pd.concat([histo_auc[["date", "auc"]], df_auc]).reset_index(drop=True)

                else:
                    warnings.warn(f"{self.datadrift_file} did not exist and is created. ")

                try:
                    df_auc.to_csv(self.datadrift_file)
                except OSError as error:
                    raise OSError("Can't save to csv the AUC metrics, error : " + str(error))
        return df_auc

    def add_data_modeldrift(
        self, dataset, metric="performance", reference_columns=[], year_col="annee", month_col="mois"
    ):
        """
        When method drift is specified, It will display in the report
        the several plots from a dataframe to analyse drift model from the deployed model.
        Each plot will represent one possible computed metric according
        to its groups. (grouped by date(year-month), reference_columns).

        Parameters
        ----------
        df : pd.DataFrame
            The Dataframe with all the computed metrics.
        metric: str, (default: 'performance')
            The column name of the metric computed
        reference_columns: list, (default: [])
            the column names to use for aggregation with the Date computed
        year_col: str, (default: 'annee')
            The column name of the year where the metric has been computed
        month_col: str, (default: 'mois')
            The column name of the month where the metric has been computed
        """
        try:
            df_modeldrift = dataset.copy()
            df_modeldrift[month_col] = df_modeldrift[month_col].apply(lambda row: str(row).split(".")[0])
            df_modeldrift["Date"] = (
                "01/" + df_modeldrift[month_col] + "/" + df_modeldrift[year_col].astype("int64").astype(str)
            )
            df_modeldrift["Date"] = pd.to_datetime(df_modeldrift["Date"], format="%d/%m/%Y")
            df_aggregate = pd.DataFrame(
                df_modeldrift.groupby(["Date"] + reference_columns)[metric].mean()
            ).reset_index()
            df_aggregate["Date"] = pd.to_datetime(df_aggregate["Date"]).dt.strftime("%d/%m/%Y")
            self.data_modeldrift = df_aggregate

        except BaseException as error:
            raise Exception(
                """
            The df specified in the method doesn't allow us to aggregate it for the report.
            - Error -
                            """
                + str(error)
            )

    def _compute_datadrift_stat_test(self, max_size=50000, categ_max=20):
        """
        calculates all statistical tests to analyze the drift of each feature

        Parameters
        ----------
        max_size : int
            Sets the maximum number of rows. If the datasets are larger there is sampling
        categ_max: int
            Maximum number of values ​​per feature to apply the chi square test

        Returns :
        -------
        dict :
            keys - features
            values - dict containing testname, statistic, pvalue
        """
        # sampling
        baseline = self.df_baseline.sample(n=max_size) if self.df_baseline.shape[0] > max_size else self.df_baseline
        current = self.df_current.sample(n=max_size) if self.df_current.shape[0] > max_size else self.df_current
        test_results = {}

        # compute test for each feature
        for features, count in self.xpl.features_desc.items():
            if current[features].dtypes.kind == "O" and count <= categ_max:
                test = chisq_test(current[features].to_numpy(), baseline[features].to_numpy())
            else:
                test = ksmirnov_test(current[features].to_numpy(), baseline[features].to_numpy())
            test_results[features] = test

        return pd.DataFrame.from_dict(test_results, orient="index")

    def define_style(self, palette_name=None, colors_dict=None):
        """
        the define_style function is a function that uses a palette or a dict
        to define the different styles used in the different outputs
        of eurybia

        Parameters
        ----------
        palette_name : str (default: 'eurybia')
            Name of the palette used for the colors of the report (refer to style folder).
        colors_dict: dict
            Dict of the colors used in the different plots
        """
        if palette_name is None and colors_dict is None:
            raise ValueError("At least one of palette_name or colors_dict parameters must be defined")
        new_palette_name = palette_name or self.palette_name
        new_colors_dict = copy.deepcopy(select_palette(colors_loading(), new_palette_name))
        if colors_dict is not None:
            new_colors_dict.update(colors_dict)
        self.colors_dict.update(new_colors_dict)
        self.plot.define_style_attributes(colors_dict=self.colors_dict)
        self.xpl.define_style(colors_dict=self.colors_dict)

    def save(self, path):
        """
        Save method allows user to save SmartDrift object on disk
        using a pickle file.
        Save method can be useful: you don't have to recompile to display
        results later

        Parameters
        ----------
        path : str
            File path to store the pickle file

        Example
        --------
        >>> smartdrift.save('path_to_pkl/smartdrift.pkl')
        """
        dict_to_save = {}
        for att in self.__dict__.keys():
            if isinstance(getattr(self, att), (list, dict, pd.DataFrame, pd.Series, type(None), bool, float)):
                dict_to_save.update({att: getattr(self, att)})
            elif isinstance(getattr(self, att), SmartExplainer):
                smartexplainer_dict = {}
                for att_xpl in self.xpl.__dict__.keys():
                    if isinstance(
                        getattr(self.xpl, att_xpl), (list, dict, pd.DataFrame, pd.Series, type(None), bool)
                    ) or att_xpl in ["model", "preprocessing", "postprocessing"]:
                        smartexplainer_dict.update({att_xpl: getattr(self.xpl, att_xpl)})
                dict_to_save.update({att: smartexplainer_dict})
        save_pickle(dict_to_save, path)
